"""Autograder runner for Python assignments"""

import json
import nbformat as nbf
import os

from glob import glob

from .abstract_runner import AbstractLanguageRunner
from ..utils import OtterRuntimeError, print_output
from ....check.logs import Log
from ....check.notebook import OTTER_LOG_FILENAME
from ....execute import grade_notebook
from ....export import export_notebook
from ....plugins import PluginCollection
from ....utils import chdir, format_full_width


class PythonRunner(AbstractLanguageRunner):

    def prepare_files(self):
        """
        Copies tests and support files needed for running the autograder.

        Also creates ``__init__.py`` files to allow for relative imports.
        """
        super().prepare_files()

        # create __init__.py files
        open("__init__.py", "a").close()
        open("submission/__init__.py", "a").close()

    def validate_submission(self, submission_path: str):
        if os.path.splitext(submission_path)[1] == ".ipynb":
            nb = nbf.read(submission_path, as_version=nbf.NO_CONVERT)
            assignment_name = self.get_notebook_assignment_name(nb)
            self.validate_assignment_name(assignment_name)

    def resolve_submission_path(self) -> str:
        nbs = glob("*.ipynb")

        if len(nbs) > 1:
            raise OtterRuntimeError("More than one .ipynb file found in submission")

        if len(nbs) == 1:
            subm_path = nbs[0]

        else:
            pys = glob("*.py")
            pys = list(filter(lambda f: f != "__init__.py", pys))
            if len(pys) > 1:
                raise OtterRuntimeError("More than one Python file found in submission")

            elif len(pys) == 1:
                subm_path = pys[0]

            else:
                raise OtterRuntimeError("No gradable files found in submission")

        return subm_path

    def write_pdf(self, submission_path: str) -> str:
        pdf_path = os.path.splitext(submission_path)[0] + ".pdf"
        export_notebook(
            submission_path,
            dest=pdf_path,
            filtering=self.ag_config.filtering,
            pagebreaks=self.ag_config.pagebreaks,
            exporter_type="html" if self.ag_config.pdf_via_html else "latex",
        )

        return pdf_path

    def run(self):
        os.environ["PATH"] = f"{self.ag_config.miniconda_path}/bin:" + os.environ.get("PATH", "")

        with chdir("./submission"):

            subm_path = self.resolve_submission_path()
            self.validate_submission(subm_path)

            # load plugins
            plugins = self.ag_config.plugins

            if plugins:
                with open("../submission_metadata.json", encoding="utf-8") as f:
                    submission_metadata = json.load(f)

                plugin_collection = PluginCollection(
                    plugins, os.path.abspath(subm_path), submission_metadata
                )

            else:
                plugin_collection = None

            if plugin_collection:
                plugin_collection.run("before_grading", self.ag_config)

            pdf_error = None
            if self.ag_config.token is not None or self.ag_config.pdf:
                pdf_error = self.write_and_maybe_submit_pdf(subm_path)

            self.sanitize_tokens()

            if os.path.isfile(OTTER_LOG_FILENAME):
                try:
                    log = Log.from_file(OTTER_LOG_FILENAME, ascending=False)

                except Exception as e:
                    if self.ag_config.grade_from_log:
                        raise e

                    else:
                        print_output(f"Could not deserialize the log due to an error:\n{e}")
                        log = None

            else:
                if self.ag_config.grade_from_log:
                    raise OtterRuntimeError("Grade from log indicated but log not found")

                log = None

            scores = grade_notebook(
                subm_path,
                tests_glob=glob("./tests/*.py"),
                cwd=os.getcwd(),
                test_dir="./tests",
                ignore_errors=not self.ag_config.debug,
                seed=self.ag_config.seed,
                seed_variable=self.ag_config.seed_variable,
                log=log if self.ag_config.grade_from_log else None,
                variables=self.ag_config.serialized_variables,
                plugin_collection=plugin_collection,
                script=os.path.splitext(subm_path)[1] == ".py",
                force_python3_kernel=not self.ag_config.otter_run,
            )

            if pdf_error:
                scores.set_pdf_error(pdf_error)

            # verify the scores against the log
            if self.ag_config.print_summary:
                print_output("\n\n\n\n", end="")
                s = format_full_width("-", mid_text="GRADING SUMMARY")
                print_output(s)
                print_output()
                if log is not None:
                    try:
                        discrepancies = scores.verify_against_log(log)
                        if self.ag_config.print_summary:
                            if not discrepancies:
                                print_output(
                                    "No discrepancies found while verifying scores against the log."
                                )
                            else:
                                for d in discrepancies:
                                    print_output(d)

                    except BaseException as e:
                        print_output(
                            f"Error encountered while trying to verify scores with log:\n{e}"
                        )

                else:
                    print_output("No log found with which to verify student scores.")

            if plugin_collection:
                report = plugin_collection.generate_report()
                if report.strip():
                    print_output("\n\n" + report)

        return scores



"""A config definition for Otter's configurations stored in the notebook metadata"""

import fica
import nbformat as nbf

from typing import Any, Optional

from .utils import NOTEBOOK_METADATA_KEY


OK_FORMAT_VARNAME = "OK_FORMAT"


class NBMetadataConfig(fica.Config):
    """
    The configurations stored in a notebook's metadata.

    This object defines the structure of the configurations that Otter can store in a notebook
    file's metadata. These configurations are parsed from ``nb["metadata"]["otter"]``.
    """

    assignment_name: Optional[str] = fica.Key(
        type_=str,
        allow_none=True,
    )
    """the name of the assignment"""

    tests: dict[str, dict[str, Any]] = fica.Key(
        type_=dict,
        # This value should never be None unless the value specified in the notebook metadata is
        # null. It is here as a patch for fica, which doesn't handle type-checking a factory
        # correctly.
        # TODO: remove allow_none=True once fica fixes this bug (chrispyles/fica#33)
        allow_none=True,
        factory=lambda: {},
    )
    """test cases keyed by question name"""

    ok_format: bool = fica.Key(
        name=OK_FORMAT_VARNAME,
        type_=bool,
        allow_none=True,
    )
    """whether the tests in ``tests`` are in the OK format"""

    require_no_pdf_confirmation: Optional[bool] = fica.Key(
        type_=bool,
        allow_none=True,
    )
    """whether to require students to acknowledge that a PDF couldn't be generated during export"""

    export_pdf_failure_message: Optional[str] = fica.Key(
        type_=str,
        allow_none=True,
    )
    """a custom message to display to students when ACKing the lack of a PDF"""

    def __init__(
        self,
        user_config: Optional[dict[str, Any]] = None,
        documentation_mode: bool = False,
        require_valid_keys: bool = False,
    ) -> None:
        if user_config is None:
            user_config = {}
        if ("tests" in user_config) != (OK_FORMAT_VARNAME in user_config):
            raise ValueError(f"{OK_FORMAT_VARNAME} must be specified with tests")

        super().__init__(user_config, documentation_mode, require_valid_keys)

    @classmethod
    def from_notebook(cls, nb: nbf.NotebookNode) -> "NBMetadataConfig":
        """
        Create a :py:class:`NBMetadataConfig` from a notebook object.
        """
        return cls(nb.get("metadata", {}).get(NOTEBOOK_METADATA_KEY, {}))



"""Output directory creation for Otter Assign"""

import nbformat
import os
import pathlib
import shutil
import tempfile

from .assignment import Assignment
from .notebook_transformer import NotebookTransformer, TransformedNotebookContainer
from .r_adapter import rmarkdown_converter
from .r_adapter.tests_manager import RAssignmentTestsManager
from .tests_manager import AssignmentTestsManager
from .utils import get_notebook_language
from .. import logging
from ..utils import NBFORMAT_VERSION


LOGGER = logging.get_logger(__name__)


def write_output_dir(
    transformed_nb: TransformedNotebookContainer,
    output_dir: pathlib.Path,
    assignment: Assignment,
    sanitize: bool,
):
    """
    Write an output directory (either for the autograder or student, as indicated by ``sanitize``).

    Args:
        transformed_nb (``otter.assign.notebook_transformed.TransformedNotebookContainer``): the
            transformed notebook
        output_dir (``pathlib.Path``): the path to the output directory being written (assumed to
            already exist)
        assignment (``otter.assign.assignment.Assignment``): the assignment config
        sanitize (``bool``): whether to sanitize the output (by removing solutions and hidden tests)
    """
    output_path = output_dir / assignment.notebook_basename
    tests_dir = output_dir / "tests"
    if assignment.tests.files:
        os.makedirs(tests_dir, exist_ok=True)

    if not sanitize:
        if assignment.requirements:
            output_fn = ("requirements.txt", "requirements.R")[assignment.is_r]
            if isinstance(assignment.requirements, list):
                with open(str(output_dir / output_fn), "w+") as f:
                    f.write("\n".join(assignment.requirements))
                assignment.requirements = str(output_dir / output_fn)
            else:
                shutil.copy(assignment.requirements, str(output_dir / output_fn))

        if assignment.environment:
            output_fn = "environment.yml"
            shutil.copy(assignment.environment, str(output_dir / output_fn))

        if assignment.student_files:
            for file in assignment.student_files:
                # TODO: dedupe this logic with handling of autograder_files utils.py and logic below
                # for assignment.files
                if not os.path.isfile(file):
                    raise FileNotFoundError(f"{file} is not a file")
                if str(assignment.master.parent) not in os.path.abspath(file):
                    raise ValueError(
                        f"{file} is not in a subdirectory of the master notebook directory"
                    )
                file_path = pathlib.Path(file).resolve()
                rel_path = file_path.parent.relative_to(assignment.master.parent)
                os.makedirs(output_dir / rel_path, exist_ok=True)
                shutil.copy(file, str(output_dir / rel_path))

    # write tests
    transformed_nb.write_tests(str(tests_dir), not sanitize, assignment.tests.files)

    # write a temp dir for otter generate tests
    if not sanitize:
        assignment.generate_tests_dir = tempfile.mkdtemp()
        transformed_nb.write_tests(assignment.generate_tests_dir, True, True)

    transformed_nb.write_transformed_nb(output_path, sanitize)

    # copy files
    for file in assignment.files:

        # if a directory, copy the entire dir
        if os.path.isdir(file):
            shutil.copytree(file, str(output_dir / os.path.basename(file)))

        else:
            # check that file is in subdir
            if str(assignment.master.parent) not in os.path.abspath(file):
                raise ValueError(
                    f"{file} is not in a subdirectory of the master notebook directory"
                )
            file_path = pathlib.Path(file).resolve()
            rel_path = file_path.parent.relative_to(assignment.master.parent)
            os.makedirs(output_dir / rel_path, exist_ok=True)
            shutil.copy(file, str(output_dir / rel_path))


def write_output_directories(assignment: Assignment):
    """
    Process a master notebook and write the results to the output directories.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    if assignment.is_rmd:
        nb = rmarkdown_converter.read_as_notebook(assignment.master)
    else:
        nb = nbformat.read(assignment.master, as_version=NBFORMAT_VERSION)

    if assignment.lang is None:
        assignment.lang = get_notebook_language(nb)

    tests_mgr = (RAssignmentTestsManager if assignment.is_r else AssignmentTestsManager)(assignment)
    nb_transformer = NotebookTransformer(assignment, tests_mgr)
    transformed_nb = nb_transformer.transform_notebook(nb)

    # update assignment.tests.files for R notebooks
    assignment.tests.files |= assignment.is_r

    # force test files if a test URL prefix is provided
    if assignment.tests["url_prefix"]:
        assignment.tests.files = True

    # create directories
    autograder_dir = assignment.get_ag_path()
    student_dir = assignment.get_stu_path()
    shutil.rmtree(autograder_dir, ignore_errors=True)
    shutil.rmtree(student_dir, ignore_errors=True)
    os.makedirs(autograder_dir, exist_ok=True)
    os.makedirs(student_dir, exist_ok=True)

    # populate directories
    write_output_dir(transformed_nb, autograder_dir, assignment, False)
    write_output_dir(transformed_nb, student_dir, assignment, True)

    # print assignment summary
    LOGGER.info(nb_transformer.tests_mgr.generate_assignment_summary())



"""Docker container management for Otter Grade"""

import importlib.resources
import json
import os
import pathlib
import shutil
import tempfile
import zipfile

from concurrent.futures import as_completed, ThreadPoolExecutor
from python_on_whales import docker
from textwrap import indent
from typing import Any, Optional

from . import __name__ as pkg_name
from .utils import OTTER_DOCKER_IMAGE_NAME, TimeoutException
from .. import logging
from ..run import AutograderConfig
from ..test_files import GradingResults
from ..utils import format_exception, OTTER_CONFIG_FILENAME


LOGGER = logging.get_logger(__name__)


def build_image(ag_zip_path: str, base_image: str, tag: str, config: AutograderConfig) -> str:
    """
    Creates a grading image based on the autograder zip file and attaches a tag.

    Args:
        ag_zip_path (``str``): path to the autograder zip file
        base_image (``str``): base Docker image to build from
        tag (``str``): tag to be added when creating the image
        config (``otter.run.run_autograder.autograder_config.AutograderConfig``): config overrides
            for the autograder

    Returns:
        ``str``: the tag of the newly-build Docker image
    """
    image = OTTER_DOCKER_IMAGE_NAME + ":" + tag
    dockerfile_path = str(importlib.resources.files(pkg_name) / "Dockerfile")

    LOGGER.info(f"Building image using {base_image} as base image")

    with tempfile.TemporaryDirectory() as temp_dir:
        with zipfile.ZipFile(ag_zip_path, "r") as zip_ref:
            zip_ref.extractall(temp_dir)

        # Update the otter_config.json file from the autograder zip with the provided config
        # overrides.
        config_path = pathlib.Path(temp_dir) / OTTER_CONFIG_FILENAME
        old_config = AutograderConfig()
        if config_path.exists():
            old_config = AutograderConfig(json.loads(config_path.read_text("utf-8")))

        old_config.update(config.get_user_config())
        config_path.write_text(json.dumps(old_config.get_user_config()))

        try:
            docker.build(
                temp_dir,
                build_args={"BASE_IMAGE": base_image},
                tags=[image],
                file=dockerfile_path,
                load=True,
            )
        except TypeError as e:
            raise TypeError(
                f"Docker build failed; if this is your first time seeing this error, ensure that "
                f"Docker is running on your machine.\n\nOriginal error: {e}"
            )

    return image


def launch_containers(
    ag_zip_path: str,
    submission_paths: list[str],
    num_containers: int,
    base_image: str,
    tag: str,
    config: AutograderConfig,
    **kwargs: Any,
) -> list[GradingResults]:
    """
    Grade submissions in parallel Docker containers.

    This function runs ``num_containers`` Docker containers in parallel to grade the student
    submissions in ``submissions_dir`` using the autograder configuration file at ``ag_zip_path``.
    If indicated, it copies the PDFs generated of the submissions out of their containers.

    Args:
        ag_zip_path (``str``): path to zip file used to set up container
        submission_paths (``str``): paths of submissions to be graded
        num_containers (``int``): number of containers to run in parallel
        base_image (``str``): the name of a base image to use for building Docker images
        tag (``str``): a tag to use for the ``otter-grade`` image created for this assignment
        config (``otter.run.run_autograder.autograder_config.AutograderConfig``): config overrides
            for the autograder
        **kwargs: additional kwargs passed to ``grade_submission``

    Returns:
        ``list[otter.test_files.GradingResults]``: the grades returned by each container spawned
            during grading
    """
    pool = ThreadPoolExecutor(num_containers)
    futures = []
    image = build_image(ag_zip_path, base_image, tag, config)

    for subm_path in submission_paths:
        futures += [
            pool.submit(
                grade_submission,
                submission_path=subm_path,
                image=image,
                **kwargs,
            )
        ]
    LOGGER.info(f"Notebooks to grade: {len(futures)}")
    scores = []
    for i, future in enumerate(as_completed(futures)):
        result = future.result()
        scores.append(result)
        LOGGER.info(f"{result.file} complete: {i+1}/{len(futures)}")

    LOGGER.info(f"Notebooks graded: {len(futures)}")
    return scores


def grade_submission(
    submission_path: str,
    image: str,
    no_kill: bool = False,
    pdf_dir: Optional[pathlib.Path] = None,
    timeout: Optional[int] = None,
    network: bool = True,
) -> GradingResults:
    """
    Grade a submission in a Docker container.

    If a submission times out, based on the timeout parameter or the container
    exits in an error state a ``GradingResults`` object is created by using the
    ``GradingResults.without_results`` function and returned.

    Args:
        submission_path (``str``): path to the submission to be graded
        image (``str``): a Docker image tag to be used for grading environment
        no_kill (``bool``): whether the grading containers should be kept running after
            grading finishes
        pdf_dir (``pathlib.Path``): a directory in which to put the notebook PDF, if applicable
        timeout (``int``): timeout in seconds for each container
        network (``bool``): whether to enable networking in the containers

    Returns:
        ``otter.test_files.GradingResults``: A ``GradingResults`` object containing the grading results
    """
    import dill

    temp_subm_file, temp_subm_path = tempfile.mkstemp()
    shutil.copyfile(submission_path, temp_subm_path)

    results_file, results_path = tempfile.mkstemp(suffix=".pkl")
    pdf_path = None
    if pdf_dir:
        pdf_file, pdf_path = tempfile.mkstemp(suffix=".pdf")

    try:
        nb_basename = os.path.basename(submission_path)
        nb_name = os.path.splitext(nb_basename)[0]

        volumes = [
            (temp_subm_path, f"/autograder/submission/{nb_basename}"),
            (results_path, "/autograder/results/results.pkl"),
        ]
        if pdf_dir:
            volumes.append((pdf_path, f"/autograder/submission/{nb_name}.pdf"))

        container = docker.container.create(
            image,
            command=["/autograder/run_autograder"],
            networks=["none"] if not network else [],
        )

        for local_path, container_path in volumes:
            docker.container.copy(local_path, (container, container_path))

        docker.container.start(container)

        did_time_out = False
        if timeout:
            import threading

            def kill_container():
                nonlocal did_time_out
                did_time_out = True
                docker.container.kill(container)

            timer = threading.Timer(timeout, kill_container)
            timer.start()

        container_id = container.id[:12]
        LOGGER.debug(f"Grading {submission_path} in container {container_id}...")
        LOGGER.info(f"Grading {nb_basename}")

        exit = docker.container.wait(container)

        if timeout:
            timer.cancel()

        logs = docker.container.logs(container)
        LOGGER.debug(f"Container {container_id} logs:\n{indent(logs, '    ')}")

        # Close our file handles since docker cp will delete the original file when performing the
        # copy.
        os.close(temp_subm_file)
        os.close(results_file)
        if pdf_path:
            os.close(pdf_file)

        for local_path, container_path in volumes:
            docker.container.copy((container, container_path), local_path)

        if not no_kill:
            container.remove()

        if did_time_out:
            raise TimeoutException(
                f"Executing '{submission_path}' in docker container timed out in {timeout} seconds"
            )

        if exit != 0:
            raise Exception(
                f"Executing '{submission_path}' in docker container failed! Exit code: {exit}"
            )

        with open(results_path, "rb") as f:
            scores = dill.load(f)

        if pdf_dir:
            pdf_dir.mkdir(parents=True, exist_ok=True)

            local_pdf_path = pdf_dir / f"{nb_name}.pdf"
            shutil.copy(pdf_path, local_pdf_path)

    except TimeoutException as te:
        scores = GradingResults.without_results(te)
        LOGGER.error(f'Submission "{nb_basename}" timed out during grading')
    except Exception as e:
        scores = GradingResults.without_results(e)
        LOGGER.error(
            f'An error occurred while grading "{nb_basename}":\n{indent(format_exception(e), "  > ")}'
        )

    finally:
        scores.file = nb_basename
        os.remove(results_path)
        os.remove(temp_subm_path)

    return scores



"""Assignment configurations for Otter Assign"""

import datetime as dt
import fica
import os
import pathlib
import yaml

from typing import Any, Literal, Optional, Union

from ..logging import Loggable
from ..run import AutograderConfig


class Assignment(fica.Config, Loggable):
    """
    Configurations for the assignment.
    """

    name: Optional[str] = fica.Key(
        description="a name for the assignment (to validate that students submit to the correct "
        "autograder)",
        default=None,
    )

    config_file: Optional[str] = fica.Key(
        description="path to a file containing assignment configurations; any configurations in "
        "this file are overridden by the in-notebook config",
        default=None,
    )

    requirements: Optional[str] = fica.Key(
        description="the path to a requirements.txt file or a list of packages",
        default=None,
    )

    overwrite_requirements: bool = fica.Key(
        description="whether to overwrite Otter's default requirement.txt in Otter Generate",
        default=False,
    )

    environment: Optional[str] = fica.Key(
        description="the path to a conda environment.yml file",
        default=None,
    )

    run_tests: bool = fica.Key(
        description="whether to run the assignment tests against the autograder notebook",
        default=True,
    )

    solutions_pdf: Union[bool, Literal["filtered"]] = fica.Key(
        description="whether to generate a PDF of the solutions notebook",
        default=False,
    )

    template_pdf: bool = fica.Key(
        description="whether to generate a filtered Gradescope assignment template PDF",
        default=False,
    )

    init_cell: bool = fica.Key(
        description="whether to include an Otter initialization cell in the output notebooks",
        default=True,
    )

    check_all_cell: bool = fica.Key(
        description="whether to include an Otter check-all cell in the output notebooks",
        default=False,
    )

    class ExportCellValue(fica.Config):

        instructions: str = fica.Key(
            description="additional submission instructions to include in the export cell",
            default="",
        )

        pdf: bool = fica.Key(
            description="whether to include a PDF of the notebook in the generated zip file",
            default=True,
        )

        filtering: bool = fica.Key(
            description="whether the generated PDF should be filtered",
            default=True,
        )

        force_save: bool = fica.Key(
            description="whether to force-save the notebook with JavaScript (only works in "
            "classic notebook)",
            default=False,
        )

        run_tests: bool = fica.Key(
            description="whether to run student submissions against local tests during export",
            default=True,
        )

        files: list[str] = fica.Key(
            description="a list of other files to include in the student submissions' zip file",
            factory=lambda: [],
        )

        class RequireNoPDFAckValue(fica.Config):

            message: str = fica.Key(
                description="a message to show to students if a PDF is meant to be included in "
                "the submission but cannot be generated",
            )

        require_no_pdf_ack: Union[bool, RequireNoPDFAckValue] = fica.Key(
            description="whether to require students to acknowledge that a PDF could not be "
            "created if one is meant to be included in the submission zip file",
            default=False,
            subkey_container=RequireNoPDFAckValue,
        )

        ignore_log: bool = fica.Key(
            description="whether to exclude the .OTTER_LOG file from the submission zip file",
            default=False,
            type_=bool,
        )

    export_cell: ExportCellValue = fica.Key(
        description="whether to include an Otter export cell in the output notebooks",
        subkey_container=ExportCellValue,
    )

    class SeedValue(fica.Config):

        variable: Optional[str] = fica.Key(
            description="a variable name to override with the autograder seed during grading",
            default=None,
        )

        autograder_value: Optional[int] = fica.Key(
            description="the value of the autograder seed",
            default=None,
        )

        student_value: Optional[int] = fica.Key(
            description="the value of the student seed",
            default=None,
        )

    seed: SeedValue = fica.Key(
        description="intercell seeding configurations",
        default=None,
        subkey_container=SeedValue,
    )

    generate: AutograderConfig = fica.Key(
        description="grading configurations to be passed to Otter Generate as an "
        "otter_config.json",
        subkey_container=AutograderConfig,
    )

    save_environment: bool = fica.Key(
        description="whether to save the student's environment in the log",
        default=False,
    )

    ignore_modules: list[str] = fica.Key(
        description="a list of modules to ignore variables from during environment serialization",
        factory=lambda: [],
    )

    files: list[str] = fica.Key(
        description="a list of other files to include in the output directories and autograder",
        factory=lambda: [],
    )

    autograder_files: list[str] = fica.Key(
        description="a list of other files only to include in the autograder",
        factory=lambda: [],
    )

    student_files: list[str] = fica.Key(
        description="a list of files that will be generated by students and thus should not be included in the autograder configuration zip file",
        factory=lambda: [],
        type_=list,
    )

    plugins: list[str] = fica.Key(
        description="a list of plugin names and configurations",
        factory=lambda: [],
    )

    class TestsValue(fica.Config):

        files: bool = fica.Key(
            description="whether to store tests in separate files, instead of the notebook "
            "metadata",
            default=False,
        )

        ok_format: bool = fica.Key(
            description="whether the test cases are in OK-format (instead of the exception-based "
            "format)",
            default=True,
        )

        url_prefix: Optional[str] = fica.Key(
            description="a URL prefix for where test files can be found for student use",
            default=None,
        )

    tests: TestsValue = fica.Key(
        description="information about the structure and storage of tests",
        subkey_container=TestsValue,
        enforce_subkeys=True,
    )

    show_question_points: bool = fica.Key(
        description="whether to add the question point values to the last cell of each question",
        default=False,
    )

    runs_on: str = fica.Key(
        description="the interpreter this notebook will be run on if different from the "
        "default interpreter (one of {'default', 'colab', 'jupyterlite'})",
        default="default",
        validator=fica.validators.choice(["default", "colab", "jupyterlite"]),
    )

    python_version: Optional[Union[str, int, float]] = fica.Key(
        description="the version of Python to use in the grading image (must be 3.9+)",
        default=None,
    )

    channel_priority_strict: bool = fica.Key(
        description="whether to set conda's channel_priority config to strict in the setup.sh file",
        default=True,
    )

    exclude_conda_defaults: bool = fica.Key(
        description="whether to exclude conda's defaults channel in the generated environment.yml file",
        default=False,
    )

    lang: Optional[str] = None
    """the language of the assignment"""

    master: pathlib.Path = None
    """the path to the master notebook"""

    result: pathlib.Path = None
    """the path to the output directory"""

    seed_required: bool = False
    """whether a seeding configuration is required for Otter Generate"""

    generate_tests_dir: Optional[str] = None
    """the path to a directory of test files for Otter Generate"""

    _ag_zip_name: Optional[str] = None
    """
    the file name for the autograder zip file; this value is generated the first time it is accessed
    since it contains a timestamp
    """

    def __init__(self, user_config: Optional[dict[str, Any]] = None, **kwargs: Any) -> None:
        if user_config is None:
            user_config = {}
        self._logger.debug(f"Initializing with config: {user_config}")
        super().__init__(user_config, **kwargs)

        # convert true values masking subkey contains to those containers
        if self.generate is True:
            self.generate = AutograderConfig()
        if self.export_cell is True:
            self.export_cell = type(self).ExportCellValue()

    def update(self, user_config: dict[str, Any]):
        self._logger.debug(f"Updating config: {user_config}")
        ret = super().update(user_config)
        if self.generate is True:
            self.generate = AutograderConfig()
        if self.export_cell is True:
            self.export_cell = type(self).ExportCellValue()
        return ret

    @property
    def is_r(self):
        """
        Whether the language of the assignment is R
        """
        return self.lang == "r"

    @property
    def is_python(self):
        """
        Whether the language of the assignment is Python
        """
        return self.lang == "python"

    @property
    def is_rmd(self):
        """
        Whether the input file is an RMarkdown document
        """
        return self.master.suffix.lower() == ".rmd"

    def get_otter_config(self):
        """
        Get the contents of ``otter_config.json`` for this assignment.

        Returns:
            ``dict[str, object]``: the ``otter_config.json`` file as a ``dict``
        """
        if self.is_r:
            self.generate.lang = "r"

        if self.name:
            self.generate.assignment_name = self.name

        return self.generate.get_user_config()

    @property
    def notebook_basename(self):
        """the basename of the notebook"""
        return os.path.basename(str(self.master))

    @property
    def ag_notebook_path(self):
        """the path to the autograder notebook"""
        return self.get_ag_path(self.notebook_basename)

    @property
    def ag_zip_name(self):
        """the file name of the autograder zip file"""
        if self._ag_zip_name is None:
            timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
            self._ag_zip_name = f"{self.master.stem}-autograder_{timestamp}.zip"
        return self._ag_zip_name

    @property
    def ag_zip_path(self):
        """the path to the autograder zip file"""
        return self.get_ag_path(self.ag_zip_name)

    def get_ag_path(self, path: Union[str, pathlib.Path] = "") -> pathlib.Path:
        """
        Get the path to the autograder output directory or a file in that directory.

        Args:
            path (``str | pathlib.Path``): a path to append to the autograder output directory path

        Returns:
            ``pathlib.Path``: the path to the autograder directory or the specified file within it
        """
        return self.result / "autograder" / path

    def get_stu_path(self, path: Union[str, pathlib.Path] = "") -> pathlib.Path:
        """
        Get the path to the student output directory or a file in that directory.

        Args:
            path (``str | pathlib.Path``): a path to append to the student output directory path

        Returns:
            ``pathlib.Path``: the path to the student directory or the specified file within it
        """
        return self.result / "student" / path

    def get_python_version(self) -> Optional[str]:
        """
        Returns the Python version indicated as a string (to avoid issues with YAML interpreting it
        as a number) if one is present.

        Returns:
            ``str | None``: the version string or ``None`` if none is present
        """
        return str(self.python_version) if self.python_version is not None else None

    def load_config_file(self, config_file: str):
        """
        Update the values in this config using the values in the specified file.

        Args:
            config_file (``str``): the configuration file to read
        """
        self._logger.info(f"Reading assignment config file {config_file}")

        with open(config_file) as f:
            config = yaml.full_load(f.read())

        if not isinstance(config, dict):
            raise TypeError("Configuration files did not produce a dictionary")

        self.update({**config, "config_file": config_file})



"""Solution removal for Otter Assign for R notebooks"""

import re


BLOCK_PROMPT = "# YOUR CODE HERE"


solution_assignment_regex = re.compile(r"(\s*[\w. $()]*(=|<-))(.*) ?# ?SOLUTION")


def solution_assignment_sub(match: re.Match[str]) -> str:
    """
    Substitutes the first matching group  with `` NULL # YOUR CODE HERE``
    """
    prefix = match.group(1)
    return prefix + " NULL # YOUR CODE HERE"


solution_line_regex = re.compile(r"(\s*).* ?# ?SOLUTION")


def solution_line_sub(match: re.Match[str]) -> str:
    """
    Substitutes the first matching group  with ``# YOUR CODE HERE``
    """
    prefix = match.group(1)
    return prefix + "# YOUR CODE HERE"


SUBSTITUTIONS = [
    (solution_assignment_regex, solution_assignment_sub),
    (solution_line_regex, solution_line_sub),
]



"""Autograder runner for R assignments"""

import copy
import nbformat as nbf
import os
import re
import tempfile
import yaml

from glob import glob
from nbconvert.exporters import ScriptExporter
from rpy2.robjects.packages import importr

from .abstract_runner import AbstractLanguageRunner
from ..utils import OtterRuntimeError
from ....export import export_notebook
from ....test_files import GradingResults
from ....utils import chdir, get_source, knit_rmd_file, NBFORMAT_VERSION


R_PACKAGES = {
    "knitr": importr("knitr"),
    "ottr": importr("ottr"),
}

RMD_YAML_REGEX = r"^\n*---\n([\s\S]+?)\n---"


class RRunner(AbstractLanguageRunner):

    subm_path_deletion_required = False
    """whether the submission path needs to be deleted (because it was created with tempfile)"""

    def validate_submission(self, submission_path: str):
        assignment_name = False
        ext = os.path.splitext(submission_path)[1].lower()
        if ext == ".ipynb":
            nb = nbf.read(submission_path, as_version=nbf.NO_CONVERT)
            assignment_name = self.get_notebook_assignment_name(nb)

        elif ext == ".rmd":
            assignment_name = None
            with open(submission_path) as f:
                rmd = f.read()
            config = re.match(RMD_YAML_REGEX, rmd)
            if config:
                config = config.group(1)
                assignment_name = yaml.full_load(config).get("assignment_name", None)

        if assignment_name is not False:
            self.validate_assignment_name(assignment_name)

    def filter_cells_with_syntax_errors(self, nb: nbf.NotebookNode) -> nbf.NotebookNode:
        """
        Filter out cells in an R notebook with syntax errors.
        """
        new_cells = []
        for cell in nb["cells"]:
            if cell["cell_type"] == "code":
                source = "\n".join(get_source(cell))
                valid_syntax = R_PACKAGES["ottr"].valid_syntax(source)[0]
                if valid_syntax:
                    new_cells.append(cell)
        nb = copy.deepcopy(nb)
        nb["cells"] = new_cells
        return nb

    def add_seeds_to_rmd_file(self, rmd_path: str):
        """
        Add intercell seeding to an Rmd file.
        """
        with open(rmd_path) as f:
            rmd = f.read()

        lines = rmd.split("\n")
        insertions = []
        for i, line in enumerate(lines):
            if line.startswith("```{r"):
                insertions.append(i)

        seed = f"set.seed({self.ag_config.seed})"
        if self.ag_config.seed_variable:
            seed = f"{self.ag_config.seed_variable} = {self.ag_config.seed}"

        for i in insertions[::-1]:
            lines.insert(i + 1, seed)

        with open(rmd_path, "w") as f:
            f.write("\n".join(lines))

    def add_seed_to_script(self, script_path: str):
        """
        Add a line calling ``set.seed`` to the top of the R script at the specified path.
        """
        with open(script_path) as f:
            script = f.read()

        script = f"set.seed({self.ag_config.seed})\n" + script

        with open(script_path, "w") as f:
            f.write(script)

    def resolve_submission_path(self) -> str:
        # create a temporary file at which to write a script if necessary
        _, script_path = tempfile.mkstemp(suffix=".R")

        # convert IPYNB files to Rmd files
        nbs = glob("*.ipynb")
        if len(nbs) > 1:
            raise OtterRuntimeError("More than one IPYNB file found in submission")

        elif len(nbs) == 1:
            nb_path = nbs[0]
            self.validate_submission(nb_path)
            nb = nbf.read(nb_path, as_version=NBFORMAT_VERSION)
            nb = self.filter_cells_with_syntax_errors(nb)

            # create the R script
            script, _ = ScriptExporter().from_notebook_node(nb)
            with open(script_path, "w") as f:
                f.write(script)

            self.subm_path_deletion_required = True
            return script_path

        # convert Rmd files to R files
        rmds = glob("*.Rmd")
        if len(rmds) > 1:
            raise OtterRuntimeError("More than one Rmd file found in submission")

        elif len(rmds) == 1:
            rmd_path = rmds[0]

            self.validate_submission(rmd_path)

            # add seeds
            if self.ag_config.seed is not None:
                self.add_seeds_to_rmd_file(rmd_path)

            # create the R script
            rmd_path = os.path.abspath(rmd_path)
            R_PACKAGES["knitr"].purl(rmd_path, script_path)

            self.subm_path_deletion_required = True
            return script_path

        os.remove(script_path)

        # get the R script
        scripts = glob("*.[Rr]")
        if len(scripts) > 1:
            raise OtterRuntimeError("More than one R script found in submission")

        elif len(scripts) == 0:
            raise OtterRuntimeError("No gradable files found in submission")

        if self.ag_config.seed is not None:
            self.add_seed_to_script(scripts[0])

        return scripts[0]

    def write_pdf(self, submission_path: str) -> str:
        # NOTE: this method ignores the submission_path argument, and instead resolves it again
        # manually
        nbs = glob("*.ipynb")
        if nbs:
            subm_path = nbs[0]
            ipynb = True

        else:
            rmds = glob("*.Rmd")
            if rmds:
                subm_path = rmds[0]
                ipynb = False

            else:
                raise OtterRuntimeError("Could not find a file that can be converted to a PDF")

        pdf_path = os.path.splitext(subm_path)[0] + ".pdf"
        if ipynb:
            export_notebook(
                subm_path,
                dest=pdf_path,
                filtering=self.ag_config.filtering,
                pagebreaks=self.ag_config.pagebreaks,
                exporter_type="html" if self.ag_config.pdf_via_html else "latex",
            )

        else:
            knit_rmd_file(subm_path, pdf_path)

        return pdf_path

    def run(self):
        os.environ["PATH"] = f"{self.ag_config.miniconda_path}/bin:" + os.environ.get("PATH", "")

        with chdir("./submission"):
            pdf_error = None
            if self.ag_config.token is not None or self.ag_config.pdf:
                pdf_error = self.write_and_maybe_submit_pdf(None)

            self.sanitize_tokens()

            subm_path = self.resolve_submission_path()
            output = R_PACKAGES["ottr"].run_autograder(
                subm_path, ignore_errors=not self.ag_config.debug, test_dir="./tests"
            )[0]
            scores = GradingResults.from_ottr_json(output)

            if pdf_error:
                scores.set_pdf_error(pdf_error)

        # delete the script if necessary
        if self.subm_path_deletion_required:
            os.remove(subm_path)
            self.subm_path_deletion_required = False

        return scores



"""PDF via LaTeX exporter"""

import os
import warnings

from textwrap import indent
from typing import Any

from .base_exporter import BaseExporter, ExportFailedException, TEMPLATE_DIR


# nbconvert can't be used on jupyterlite, so try importing it and raise an error if it's not found
# when the exporter is used
_NBCONVERT_ERROR = None
try:
    import nbconvert
except ImportError as e:
    nbconvert = None
    _NBCONVERT_ERROR = e


class PDFViaLatexExporter(BaseExporter):
    """
    An exporter that uses nbconvert's PDF exporter to convert notebooks to PDFs via LaTeX.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update(
        {
            "save_tex": False,
            "template": "via_latex",
        }
    )

    @classmethod
    def convert_notebook(cls, nb_path: str, dest: str, *, xecjk: bool = False, **kwargs: Any):
        if nbconvert is None:
            raise _NBCONVERT_ERROR

        warnings.filterwarnings("ignore", r"invalid escape sequence '\\c'", DeprecationWarning)

        options = cls.default_options.copy()
        options.update(kwargs)

        # choose the template to allow Chinese, Japanese, and Korean characters if necessary
        if xecjk:
            options["template"] = "via_latex_xecjk"

        nb = cls.load_notebook(
            nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"]
        )

        nbconvert.TemplateExporter.extra_template_basedirs = [str(TEMPLATE_DIR)]
        orig_template_name = nbconvert.TemplateExporter.template_name
        nbconvert.TemplateExporter.template_name = options["template"]

        if options["save_tex"]:
            latex_exporter = nbconvert.LatexExporter()

        pdf_exporter = nbconvert.PDFExporter()

        try:
            if options["save_tex"]:
                latex_output = nbconvert.export(latex_exporter, nb)
                with open(os.path.splitext(dest)[0] + ".tex", "w+") as output_file:
                    output_file.write(latex_output[0])

            pdf_output = nbconvert.export(pdf_exporter, nb)
            with open(dest, "wb") as output_file:
                output_file.write(pdf_output[0])

        except nbconvert.exporters.pdf.LatexFailed as error:
            message = "There was an error generating your LaTeX; showing full error message:\n"
            message += indent(error.output, "    ")
            if xecjk:
                message += (
                    "\n\nIf the error above is related to xeCJK or fandol in LaTeX "
                    "and you don't require this functionality, try running again without "
                    "xecjk set to True or the --xecjk flag."
                )
            raise ExportFailedException(message)

        finally:
            nbconvert.TemplateExporter.template_name = orig_template_name



"""Feature toggling for Otter Assign"""

from enum import Enum
from typing import Callable

from .assignment import Assignment


_NOT_RMD: Callable[[Assignment], bool] = lambda a: not a.is_rmd
"""a function that returns ``True`` if the assignment is not an R Markdown assignment"""


class _FeatureToggleValue:
    test: Callable[[Assignment], bool]

    def __init__(self, test: Callable[[Assignment], bool]):
        self.test = test

    def is_enabled(self, assignment: Assignment):
        return self.test(assignment)


class FeatureToggle(Enum):
    """
    An enum for managing features that are enabled or disabled depending on the assignment config.
    """

    PDF_FILTERING_COMMENTS = _FeatureToggleValue(_NOT_RMD)
    """whether PDF filtering HTML comments should be inserted into the notebook"""

    EMPTY_MD_BOUNDARY_CELLS = _FeatureToggleValue(_NOT_RMD)
    """whether to use empty Markdown cells for boundaries"""



"""Gradescope token generator; original script provided by Gradescope"""

import getpass
import requests

from typing import Any, Optional


BASE_URL = "https://www.gradescope.com"


class APIClient:
    """
    Client for Gradescope's API. Logs in users and retrieves token, uploads PDF and code assignment
    submissions.

    Client code originally provided by Gradescope.

    Args:
        token (``str``): a pre-retrieved token for the client

    Attributes:
        token (``str``): the user's Gradescope token
    """

    def __init__(self, token: Optional[str] = None):
        self.session = requests.Session()
        if token is not None:
            self.token = token

    @classmethod
    def get_token(cls) -> str:
        """
        Prompts a user for their Gradescope username and password and retrieves a token for that user

        Returns:
            ``str``: the user's token
        """
        client = cls()
        email = input("Please provide the email address on your Gradescope account: ")
        password = getpass.getpass("Gradescope password: ")
        client.log_in(email, password)
        return client.token

    def post(self, *args: Any, **kwargs: Any) -> requests.Response:
        return self.session.post(*args, **kwargs)

    def log_in(self, email: str, password: str):
        url = "{base}/api/v1/user_session".format(base=BASE_URL)

        form_data = {"email": email, "password": password}
        r = self.post(url, data=form_data)
        assert r.status_code == 200, r.text

        self.token = r.json()["token"]

    def upload_pdf_submission(
        self, course_id: str, assignment_id: str, student_email: str, filename: str
    ) -> requests.Response:
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {"owner_email": student_email}
        files = {"pdf_attachment": open(filename, "rb")}

        request_headers = {"access-token": self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r

    def replace_pdf_submission(
        self, course_id: str, assignment_id: str, student_email: str, filename: str
    ) -> requests.Response:
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions/replace_pdf".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {"owner_email": student_email}
        files = {"pdf_attachment": open(filename, "rb")}

        request_headers = {"access-token": self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r

    def upload_programming_submission(
        self, course_id: str, assignment_id: str, student_email: str, filenames: list[str]
    ) -> requests.Response:
        url = "{base}/api/v1/courses/{0}/assignments/{1}/submissions".format(
            course_id, assignment_id, base=BASE_URL
        )

        form_data = {"owner_email": student_email}
        files = [("files[]", (filename, open(filename, "rb"))) for filename in filenames]

        request_headers = {"access-token": self.token}
        r = self.post(url, data=form_data, headers=request_headers, files=files)
        return r



"""Various utilities for Otter-Grader"""

import nbformat
import os
import pathlib
import random
import re
import shutil
import string
import tempfile
import traceback
import yaml

from collections.abc import Generator
from contextlib import contextmanager
from IPython.core.getipython import get_ipython
from typing import Any, Optional


NBFORMAT_VERSION = 4
"""the version of the Jupyter notebook format to use"""

NOTEBOOK_METADATA_KEY = "otter"
"""the key used for all Otter metadata added to a notebook"""

REQUIRE_CONFIRMATION_NO_PDF_EXPORT_KEY = "require_no_pdf_confirmation"
"""
the key in Otter's notebook metadata for requiring students to acknowledge that their notebook could
not be exported as a PDF before creating the submission zip file
"""

NO_PDF_EXPORT_MESSAGE_KEY = "export_pdf_failure_message"
"""
the key in Otter's notebook metadata for the message to show if a notebook cannot be exported as a
PDF
"""

OTTER_CONFIG_FILENAME = "otter_config.json"
"""the file name for the autograder config JSON file"""


@contextmanager
def hide_outputs():
    """
    Context manager for hiding outputs from ``display()`` calls. IPython handles matplotlib outputs
    specially, so those are suppressed too.
    """
    ipy = get_ipython()
    if ipy is None:
        # Not running inside ipython!
        yield
        return
    old_formatters = ipy.display_formatter.formatters
    ipy.display_formatter.formatters = {}
    try:
        yield
    finally:
        ipy.display_formatter.formatters = old_formatters


def id_generator(size: int = 6, chars: str = string.ascii_uppercase + string.digits) -> str:
    """
    Used to generate a dynamic variable name for grading functions

    This function generates a random name using the given length and character set.

    Args:
        size (``int``): length of output name
        chars (``str``): set of characters used to create function name

    Returns:
        ``str``: randomized string name for grading function
    """
    return "".join(random.choice(chars) for _ in range(size))


def get_variable_type(obj: Any) -> str:
    """
    Returns the fully-qualified type string of an object ``obj``

    Args:
        obj (any): the object in question

    Returns:
        ``str``: the fully-qualified type string
    """
    return type(obj).__module__ + "." + type(obj).__name__


def get_relpath(src: pathlib.Path, dst: pathlib.Path) -> pathlib.Path:
    """
    Returns the relative path from ``src`` to ``dst``

    Args:
        src (``pathlib.Path``): the source directory
        dst (``pathlib.Path``): the destination directory

    Returns:
        ``pathlib.Path``: the relative path
    """
    ups = 0
    while True:
        try:
            dst.relative_to(src)
            break
        except ValueError:
            src = src.parent
            ups += 1
    return pathlib.Path(("../" * ups) / dst.relative_to(src))


@contextmanager
def chdir(new_dir: "os.PathLike[Any]"):
    """
    Create a context with a different working directory, resetting the working directory on exit.

    Args:
        new_dir (path-like): the directory for the context
    """
    curr_dir = os.getcwd()
    os.chdir(new_dir)

    try:
        yield

    finally:
        os.chdir(curr_dir)


def get_source(cell: nbformat.NotebookNode) -> list[str]:
    """
    Returns the source code of a cell in a way that works for both nbformat and JSON

    Args:
        cell (``nbformat.NotebookNode``): notebook cell

    Returns:
        ``list[str]``: each line of the cell source stripped of ending line breaks
    """
    source = cell.source
    if isinstance(source, str):
        return re.split("\r?\n", source)
    elif isinstance(source, list):
        return [line.strip("\r\n") for line in source]
    raise TypeError(f"Unknown cell source type: {type(source)}")


@contextmanager
def load_default_file(
    provided_filename: Optional[str], default_filename: str, default_disabled: bool = False
) -> Generator[Optional[str]]:
    """
    Reads the contents of a file with an optional default path. If ``provided_filename`` is not specified
    and ``default_filename`` is an existing file path, the contents of ``default_filename`` are read in place
    of ``provided_filename``. The use of ``default_filename`` can be disabled by setting ``default_disabled``
    to ``True``.
    """
    if provided_filename is None and os.path.isfile(default_filename) and not default_disabled:
        provided_filename = default_filename

    if provided_filename is not None:
        if not os.path.isfile(provided_filename):
            raise FileNotFoundError(f"Could not find specified file: {provided_filename}")

        with open(provided_filename) as f:
            yield f.read()

    else:
        yield


def format_full_width(char: str, mid_text: str = "", whitespace: str = " ") -> str:
    """
    Format a character at the full terminal width. If ``mid_text`` is supplied, this text is printed
    in the middle of the terminal, surrounded by ``whitespace``.
    """
    cols, _ = shutil.get_terminal_size()

    if mid_text:
        left = cols - len(mid_text) - 2 * len(whitespace)
        if left <= 0:
            left = 2
        l, r = left // 2, left // 2
        if left % 2 == 1:
            r += 1

        out = char * l + whitespace + mid_text + whitespace + char * r

    else:
        out = char * cols

    return out


def print_full_width(char: str, mid_text: str = "", whitespace: str = " ", **kwargs: Any):
    """
    Prints a character at the full terminal width. Additional kwargs passed to ``print``.

    See ``format_full_width`` for details on the named arguments.
    """
    print(format_full_width(char, mid_text, whitespace), **kwargs)


def assert_path_exists(path_tuples: list[tuple[str, bool]]):
    """
    Ensure that a series of file paths exist and are of a specific type, or raise a ``ValueError``.

    Elements of ``path_tuples`` should be 2-tuples where the first element is a string representing
    the file path and the second element is ``True`` if the path should be a directory, ``False`` if
    it should be a file, and ``None`` if it doesn't matter.

    Args:
        path_tuples (``list[tuple[str, bool]]``): the list of paths as described above

    Raises:
        ``FileNotFoundError``: if the path does not exist or it is not of the correct type
    """
    for path, is_dir in path_tuples:
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path {path} does not exist")
        if is_dir and not os.path.isdir(path):
            raise FileNotFoundError(f"Path {path} is not a directory")
        if is_dir is False and not os.path.isfile(path):
            raise FileNotFoundError(f"Path {path} is not a file")


def knit_rmd_file(rmd_path: str, pdf_path: str):
    """
    Use ``rpy2`` and ``rmarkdown::render`` to knit an RMarkdown file to a PDF, allowing errors.

    Args:
        rmd_path (``str``): the path to the Rmd file
        pdf_path (``str``): the path at which to write the PDF
    """
    from rpy2.robjects.packages import importr

    with tempfile.NamedTemporaryFile(mode="w", suffix=".Rmd") as ntf:
        with open(rmd_path) as f:
            contents = f.read()

        contents = (
            "```{r cache = F, include = F}\nknitr::opts_chunk$set(error = TRUE)\n```\n" + contents
        )
        ntf.write(contents)
        ntf.seek(0)

        pdf_path = os.path.abspath(pdf_path)
        rmarkdown = importr("rmarkdown")
        rmarkdown.render(ntf.name, "pdf_document", pdf_path)


class _CorrectIndentationDumper(yaml.Dumper):
    def increase_indent(self, flow: bool = False, *args: Any, **kwargs: Any):
        return super().increase_indent(flow=flow, indentless=False)


def dump_yaml(o: Any, **kwargs: Any) -> str:
    """
    Dump an object to a YAML string using the ``_CorrectIndentationDumper`` dumper.

    Args:
        o (``object``): the object to dump
        **kwargs: additional keyword arguments passed to ``yaml.dump``

    Returns:
        ``str``: the YAML representation of ``o``
    """
    return yaml.dump(o, sort_keys=False, Dumper=_CorrectIndentationDumper, **kwargs)


class QuestionNotInLogException(Exception):
    """
    Exception that indicates that a specific question was not found in any entry in the log
    """


def format_exception(e: Exception) -> str:
    """
    Formats an exception for display with its traceback using the ``traceback`` module.

    Args:
        e (``Exception``): the exception to format

    Returns:
        ``str``: the formatted exception
    """
    return "".join(traceback.format_exception(type(e), e, e.__traceback__))



import ast
import json
import nbformat as nbf
import os
import tempfile

from nbconvert.exporters import PythonExporter
from nbconvert.preprocessors import Preprocessor
from textwrap import dedent
from traitlets import Bool, Dict, Instance, Integer, List, Unicode
from typing import Optional, TypeVar

from ..check.logs import Log
from ..utils import id_generator, NOTEBOOK_METADATA_KEY


T = TypeVar("T")


IGNORE_CELL_TAG = "otter_ignore"

INIT_CELL_SOURCE = """\
from otter import Notebook as {notebook_name}
{notebook_name}.init_grading_mode("{test_dir}")

# set up Otter's logging
import logging
from otter import logging
logging.set_level(logging.DEBUG)
logging.send_logs("{logging_server_host}", {logging_server_port})
"""

EXPORT_CELL_SOURCE = """\
from otter.execute import Checker
for {tests_loop_var} in {tests_glob_json}:
    Checker.check_if_not_already_checked({tests_loop_var})

from otter.test_files import GradingResults
results = GradingResults(Checker.get_results())

import pickle
with open("{results_path}", "wb") as f:
    pickle.dump(results, f)
"""


class GradingPreprocessor(Preprocessor):

    cwd = Unicode(allow_none=True).tag(config=True)

    test_dir = Unicode().tag(config=True)

    tests_glob = List(Unicode()).tag(config=True)

    results_path = Unicode().tag(config=True)

    seed = Integer(allow_none=True).tag(config=True)

    seed_variable = Unicode(allow_none=True).tag(config=True)

    otter_log: Optional[Log] = Instance(Log, allow_none=True).tag(config=True)

    variables = Dict(value_trait=Unicode(), key_trait=Unicode(), allow_none=True).tag(config=True)

    _notebook_name: str = ""

    _log_temp_file: Optional[tuple[int, str]] = None

    logging_server_host = Unicode().tag(config=True)

    logging_server_port = Integer().tag(config=True)

    force_python3_kernel = Bool().tag(config=True)

    @property
    def from_log(self):
        return self.otter_log is not None

    def preprocess(self, nb: nbf.NotebookNode, resources: T = None) -> tuple[nbf.NotebookNode, T]:
        self._notebook_name = f"notebook_{id_generator()}"
        self.filter_ignored_cells(nb)
        self.logging_transform(nb)
        self.add_checks(nb)
        self.add_seeds(nb)
        self.add_cwd_to_path(nb)
        self.update_kernel(nb)
        self.add_init_and_export_cells(nb)  # this should always be the last call
        return nb, resources

    def add_init_and_export_cells(self, nb: nbf.NotebookNode):
        nb.cells.insert(
            0,
            nbf.v4.new_code_cell(
                INIT_CELL_SOURCE.format(
                    notebook_name=self._notebook_name,
                    test_dir=self.test_dir,
                    logging_server_host=self.logging_server_host,
                    logging_server_port=self.logging_server_port,
                )
            ),
        )
        nb.cells.append(
            nbf.v4.new_code_cell(
                EXPORT_CELL_SOURCE.format(
                    tests_loop_var=f"t_{id_generator()}",
                    tests_glob_json=json.dumps(self.tests_glob),
                    # ensure that "\" is properly-escaped for Windows paths since this is going to be
                    # rendered into a string literal
                    results_path=self.results_path.replace("\\", "\\\\"),
                )
            )
        )

    def add_cwd_to_path(self, nb: nbf.NotebookNode):
        if self.cwd:
            nb.cells.insert(0, nbf.v4.new_code_cell(f'import sys\nsys.path.append(r"{self.cwd}")'))

    def add_seeds(self, nb: nbf.NotebookNode):
        if self.seed is None or self.from_log:
            return

        skip_first = False
        if self.seed_variable is None:
            skip_first = True
            np_name, rand_name = f"np_{id_generator()}", f"random_{id_generator()}"
            nb.cells.insert(
                0, nbf.v4.new_code_cell(f"import numpy as {np_name}\nimport random as {rand_name}")
            )
            do_seed = f"{np_name}.random.seed({self.seed})\n{rand_name}.seed({self.seed})"

        else:
            do_seed = f"{self.seed_variable} = {self.seed}"

        cells = nb.cells[1:] if skip_first else nb.cells
        for cell in cells:
            if cell.cell_type == "code" and not cell.source.startswith("%%"):
                cell.source = f"{do_seed}\n{cell.source}"

    def add_checks(self, nb: nbf.NotebookNode):
        if self.from_log:
            return

        new_cells = []
        for cell in nb.cells:
            new_cells.append(cell)

            config = cell.get("metadata", {}).get(NOTEBOOK_METADATA_KEY, {})
            if config.get("tests"):
                source = ""
                for test in config["tests"]:
                    source += f'{self._notebook_name}().check("{test}")\n'

                new_cells.append(nbf.v4.new_code_cell(source))

        nb.cells = new_cells

    def filter_ignored_cells(self, nb: nbf.NotebookNode):
        new_cells = []
        for cell in nb.cells:
            m = cell.get("metadata", {})
            if IGNORE_CELL_TAG in m.get("tags", []) or m.get(NOTEBOOK_METADATA_KEY, {}).get(
                "ignore", False
            ):
                continue

            new_cells.append(cell)

        nb.cells = new_cells

    def logging_transform(self, nb: nbf.NotebookNode):
        if not self.from_log:
            return

        self.filter_out_non_imports(nb)

        self._log_temp_file = tempfile.mkstemp()
        log_fn = self._log_temp_file[1]
        for e in self.otter_log.entries:
            e.flush_to_file(log_fn)

        nb.cells.append(
            nbf.v4.new_code_cell(
                dedent(
                    f"""\
            import json
            from otter import Notebook
            from otter.check.logs import Log
            from otter.utils import get_variable_type

            variables = json.loads(\"\"\"{json.dumps(self.variables)}\"\"\")
            log = Log.from_file("{log_fn}")
            logged_questions = []
            grader = Notebook()

            for entry in log.question_iterator():
                shelf = entry.unshelve(globals())

                if variables is not None:
                    for k, v in shelf.items():
                        full_type = get_variable_type(v)
                        if not (k in variables and variables[k] == full_type):
                            del shelf[k]
                            print(f"Found variable of different type than expected: {{k}}")

                globals().update(shelf)
                grader.check(entry.question)
                logged_questions.append(entry.question)

            print(f"Questions executed from log: {{', '.join(logged_questions)}}")
        """
                )
            )
        )

    def filter_out_non_imports(self, nb: nbf.NotebookNode):
        e = PythonExporter()
        ic = ImportCollector()
        tree = ast.parse(e.from_notebook_node(nb)[0])
        ic.visit(tree)
        nb.cells = [nbf.v4.new_code_cell(ic.to_script())]

    def update_kernel(self, nb: nbf.NotebookNode):
        if not self.force_python3_kernel:
            return
        nb["metadata"].get("kernelspec", {})["name"] = "python3"

    def cleanup(self):
        if self._log_temp_file:
            os.close(self._log_temp_file[0])
            os.remove(self._log_temp_file[1])


class ImportCollector(ast.NodeVisitor):
    imports = []

    def visit_Import(self, node: ast.Import):
        self.imports.append(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        self.imports.append(node)

    def to_module(self):
        return ast.Module(body=self.imports, type_ignores=[])

    def to_script(self):
        return ast.unparse(self.to_module())



"""IPython notebook PDF Exporter forked from nb2pdf and gsExport"""

import os

from typing import Any, Optional

from .exporters import ExporterType


def export_notebook(
    nb_path: str,
    dest: Optional[str] = None,
    exporter_type: Optional[ExporterType] = None,
    **kwargs: Any,
):
    """
    Exports a notebook file at ``nb_path`` to a PDF with optional filtering and pagebreaks. Accepts
    other ``kwargs`` passed to the exporter class's ``convert_notebook`` class method.

    Args:
        nb_path (``str``): path to notebook
        dest (``str | None``): path to write PDF
        exporter_type (``"html" | "latex" | None``): the type of exporter to use
        **kwargs: additional configurations passed to exporter

    Returns:
        ``str``: the path at which the PDF was written
    """
    # If nbconvert is not installed, users should be informed that it is required for Otter Export.
    try:
        import nbconvert as _
    except:
        raise ImportError("nbconvert is required for Otter Export but it could not be found")

    from .exporters import get_exporter

    if dest is not None:
        pdf_name = dest
    else:
        pdf_name = os.path.splitext(nb_path)[0] + ".pdf"

    Exporter = get_exporter(exporter_type=exporter_type)
    Exporter.convert_notebook(nb_path, pdf_name, **kwargs)

    return pdf_name


def main(
    src: str,
    *,
    dest: Optional[str] = None,
    exporter: Optional[ExporterType] = None,
    filtering: bool = False,
    pagebreaks: bool = False,
    save: bool = False,
    xecjk: bool = False,
):
    """
    Runs Otter Export

    Args:
        src (``str``): path to source notebook
        dest (``str | None``): path at which to write PDF
        exporter (``str | None``): exporter name
        filtering (``bool``): whether to filter cells using HTML comments
        pagebreaks (``bool``): whether to pagebreak between filtered regions; ignored if ``filtering``
            is ``False``
        save (``bool``): whether to save any intermediate files (e.g. ``.tex``, ``.html``)
        xecjk (``bool``): whether to use xeCJK in the LaTeX template
    """
    export_notebook(
        src,
        dest=dest,
        exporter_type=exporter,
        filtering=filtering,
        pagebreaks=pagebreaks,
        save_tex=save,
        save_html=save,
        xecjk=xecjk,
    )



"""Cell factory for Otter Assign"""

import nbformat

from .assignment import Assignment
from .feature_toggle import FeatureToggle
from .question_config import QuestionConfig
from .solutions import ANSWER_CELL_TAG
from .utils import lock


class CellFactory:
    """
    A factory for cells that make use of Otter's client package (e.g. init cell, check cell).

    All (non-static cell-generating) methods in this factory should return a
    ``list[nbformat.NotebookNode]``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """

    assignment: Assignment
    """the assignment config"""

    def __init__(self, assignment: Assignment):
        self.assignment = assignment

    def check_feature_toggle(self, feature_toggle: FeatureToggle) -> bool:
        """
        Check whether the specified feature is enabled for this assignment.

        Args:
            feature_toggle (``otter.assign.feature_toggle.FeatureToggle``): the feature

        Returns:
            ``bool``: whether the feature is enabled
        """
        return feature_toggle.value.is_enabled(self.assignment)

    def create_init_cells(self) -> list[nbformat.NotebookNode]:
        """
        Generate a cell to initialize Otter in the notebook.

        Returns:
            ``list[nbformat.NotebookNode]``: the init cell
        """
        args = ""
        if self.assignment.runs_on == "default":
            args = f'"{self.assignment.master.name}"'

        if self.assignment.tests.url_prefix:
            args += f"{', ' if args else ''}tests_url_prefix=\"{self.assignment.tests.url_prefix}\""

        contents = f"# Initialize Otter\nimport otter\ngrader = otter.Notebook({args})"
        cell = nbformat.v4.new_code_cell(contents)
        lock(cell)

        extra_cells = []
        if self.assignment.runs_on == "colab":
            pip_cell = nbformat.v4.new_code_cell("%pip install -q otter-grader")
            lock(pip_cell)
            extra_cells.append(pip_cell)

        return extra_cells + [cell]

    def create_check_cells(self, question: QuestionConfig) -> list[nbformat.NotebookNode]:
        """
        Create a cell calling ``otter.Notebook.check`` for the specified question.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``list[nbformat.NotebookNode]``: the check cell
        """
        cell = nbformat.v4.new_code_cell()
        cell.source = ['grader.check("{}")'.format(question.name)]
        lock(cell)
        return [cell]

    def create_check_all_cells(self) -> list[nbformat.NotebookNode]:
        """
        Generate a check-all cell and a Markdown cell with instructions to run all tests in the
        notebook.

        Returns:
            ``list[nbformat.NotebookNode]``: the check-all cells
        """
        instructions = nbformat.v4.new_markdown_cell()
        instructions.source = (
            "---\n\nTo double-check your work, the cell below will rerun all "
            "of the autograder tests."
        )

        check_all = nbformat.v4.new_code_cell("grader.check_all()")

        lock(instructions)
        lock(check_all)

        return [instructions, check_all]

    def create_export_cells(self) -> list[nbformat.NotebookNode]:
        """
        Generate export cells that instruct the student the run a code cell calling
        ``otter.Notebook.export`` to generate and download their submission.

        Returns:
            ``list[nbformat.NotebookNode]``: the export cells
        """
        if not self.assignment.export_cell:
            return []

        instructions = nbformat.v4.new_markdown_cell()
        instructions.source = (
            "## Submission\n\nMake sure you have run all cells in your "
            "notebook in order before running the cell below, so that all images/graphs appear "
            "in the output. The cell below will generate a zip file for you to submit."
        )

        # only include save text if force_save is false
        if not self.assignment.export_cell.force_save:
            instructions.source += " **Please save before exporting!**"

        if self.assignment.export_cell.instructions:
            instructions.source += "\n\n" + self.assignment.export_cell.instructions

        export = nbformat.v4.new_code_cell()
        source_lines = []

        # only include save text if force_save is false
        if not self.assignment.export_cell.force_save:
            source_lines.append(
                "# Save your notebook first, then run this cell to export your submission."
            )

        args = []
        if not self.assignment.export_cell.filtering:
            args += ["filtering=False"]
        if not self.assignment.export_cell.pdf:
            args += ["pdf=False"]
        if self.assignment.export_cell.force_save:
            args += ["force_save=True"]
        if self.assignment.export_cell.run_tests:
            args += ["run_tests=True"]
        if self.assignment.export_cell.ignore_log:
            args += ["ignore_log=True"]
        if self.assignment.export_cell.files or self.assignment.student_files:
            l = [*self.assignment.export_cell.files, *self.assignment.student_files]
            args += [f"files={l}"]
        source_lines.append(f"grader.export({', '.join(args)})")
        export.source = "\n".join(source_lines)

        lock(instructions)
        lock(export)

        cells = [instructions, export]
        if self.check_feature_toggle(FeatureToggle.EMPTY_MD_BOUNDARY_CELLS):
            cells.append(nbformat.v4.new_markdown_cell(" "))  # add buffer cell

        return cells

    @staticmethod
    def create_markdown_response_cell() -> nbformat.NotebookNode:
        """
        Generate a Markdown response cell with the following contents:

        .. code-block:: markdown

            _Type your answer here, replacing this text._

        Note that, unlike the other methods, this method returns a single cell rather than a list of
        cells (since it is not used in the same context).

        Returns:
            ``nbformat.NotebookNode``: the response cell
        """
        return nbformat.v4.new_markdown_cell(
            "_Type your answer here, replacing this text._", metadata={"tags": [ANSWER_CELL_TAG]}
        )



"""A Python module for validating that a submission zip file passes test cases"""

import warnings


# suppress all warnings, because otherwise they get printed to stderr which triggers a RuntimeError
# in the calling code -- see #735
warnings.simplefilter("ignore")


import argparse
import dill
import os
import shutil
import tempfile
import zipfile

from glob import glob

from ...execute import grade_notebook


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--zip-path", required=True)
    parser.add_argument("--nb-arcname", required=True)
    parser.add_argument("--tests-dir", required=True)
    parser.add_argument("--results-path", required=True)
    return parser


def main():
    args = get_parser().parse_args()

    nb_dir = tempfile.mkdtemp()

    try:
        with zipfile.ZipFile(args.zip_path, "r") as zf:
            nb_path = zf.extract(args.nb_arcname, path=nb_dir)

        results = grade_notebook(
            nb_path,
            test_dir=args.tests_dir,
            tests_glob=glob(args.tests_dir + "/*.py"),
            cwd=os.getcwd(),
        )

        with open(args.results_path, "wb") as f:
            dill.dump(results, f)

    finally:
        shutil.rmtree(nb_dir)


if __name__ == "__main__":
    main()



"""Utilities for Otter Check"""

import ipylab
import nbformat as nbf
import os
import sys
import tempfile
import time
import warnings
import wrapt

from dataclasses import dataclass
from enum import Enum
from glob import glob
from IPython.core.getipython import get_ipython
from IPython.display import display, Javascript
from ipywidgets import Button, HTML, Output, VBox
from subprocess import PIPE, run
from typing import Any, Callable, Generic, Optional, TYPE_CHECKING, TypeVar, Union

from ..nbmeta_config import NBMetadataConfig
from ..test_files import GradingResults
from ..utils import format_exception


T = TypeVar("T")


def save_notebook(filename: str, timeout: Union[int, float] = 10) -> bool:
    """
    Force-saves a Jupyter notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int | float``): number of seconds to wait for save before timing-out

    Returns
        ``bool``: whether the notebook was saved successfully
    """
    if get_ipython() is not None:
        orig_mod_time = os.path.getmtime(filename)
        start = time.time()

        # For classic notebooks
        display(
            Javascript(
                """
            if (typeof Jupyter !== 'undefined') {
                Jupyter.notebook.save_checkpoint();
            }
        """
            )
        )

        # For JupyterLab
        try:
            app = ipylab.JupyterFrontEnd()
            app.commands.execute("docmanager:save")
        except:
            pass

        while time.time() - start < timeout:
            curr_mod_time = os.path.getmtime(filename)
            if orig_mod_time < curr_mod_time and os.path.getsize(filename) > 0:
                return True

            time.sleep(0.2)

        return False

    return True


def grade_zip_file(zip_path: str, nb_arcname: str, tests_dir: str) -> GradingResults:
    """
    Grade a submission zip file in a separate process and return the resulting
    :py:class:`otter.test_files.GradingResults` object.

    Args:
        zip_path (``str``): the path to the submission zip file
        nb_arcname (``str``): the name of the notebook file in the zip file
        tests_dir (``str``): the path to the tests directory

    Returns:
        ``otter.test_files.GradingResults``: the grading results
    """
    import dill

    results_handle, results_path = tempfile.mkstemp(suffix=".pkl")

    try:
        command = [
            sys.executable,
            "-m",
            "otter.check.validate_export",
            "--zip-path",
            zip_path,
            "--nb-arcname",
            nb_arcname,
            "--tests-dir",
            tests_dir,
            "--results-path",
            results_path,
        ]

        # this environment variable is needed to fix #686
        subprocess_env = {**os.environ, "PYDEVD_DISABLE_FILE_VALIDATION": "1"}

        # run the command
        results = run(command, env=subprocess_env, stdout=PIPE, stderr=PIPE)

        print(results.stdout.decode("utf-8"))

        if results.stderr:
            warnings.warn(results.stderr.decode("utf-8"), RuntimeWarning)

        with open(results_path, "rb") as f:
            results = dill.load(f)

        return results

    finally:
        os.close(results_handle)
        os.remove(results_path)


@dataclass
class _Interpreter:
    """
    A class representing a flavor of IPython interpreter.

    Contains attributes for an importable name substring (used to check which interpreter is
    running) and a display name for error messages and the like.
    """

    check_strs: list[str]
    """list of substrings to check for in the IPython interpreter string"""

    display_name: str
    """a display name for this interpreter"""

    def running(self) -> bool:
        """
        Determine whether this interpreter is currently running by checking the string
        representation of the return value of ``IPython.get_ipython``.

        Returns:
            ``bool``: whether this interpreter is running
        """
        ipython_interp = str(get_ipython())
        return any(c in ipython_interp for c in self.check_strs)


class IPythonInterpreter(Enum):
    """
    An enum of different types of IPython interpreters.
    """

    COLAB = _Interpreter(["google.colab"], "Google Colab")
    """the Google Colab interpreter"""

    PYOLITE = _Interpreter(["pyolite.", "pyodide_kernel."], "Jupyterlite")
    """the JupyterLite interpreter"""


def incompatible_with(
    interpreter: IPythonInterpreter, throw_error: bool = True
) -> Callable[[Callable[..., T]], Callable[..., Optional[T]]]:
    """
    Create a decorator indicating that a method is incompatible with a specific interpreter.
    """

    @wrapt.decorator
    def incompatible(
        wrapped: Callable[..., T], self: "Notebook", args: tuple[Any], kwargs: dict[str, Any]
    ) -> Optional[T]:
        """
        A decorator that raises an error or performs no action (depending on ``throw_error``) if the
        wrapped function is called in an environment running on the specified interpreter.
        """
        if self.interpreter is interpreter:
            if throw_error:
                raise RuntimeError(
                    f"This method is not compatible with {interpreter.value.display_name}"
                )
            else:
                return
        return wrapped(*args, **kwargs)

    return incompatible


@wrapt.decorator
def grading_mode_disabled(
    wrapped: Callable[..., T],
    self: "Notebook",
    args: tuple[Any],
    kwargs: dict[str, Any],
) -> Optional[T]:
    """
    A decorator that returns without calling the wrapped function if the ``Notebook`` grading mode
    is enabled.
    """
    if type(self)._grading_mode:
        return
    return wrapped(*args, **kwargs)


@dataclass
class LoggedEventReturnValue(Generic[T]):
    return_value: T
    question: Optional[str] = None
    shelve_env: Optional[dict[str, Any]] = None


def logs_event(
    event_type: "EventType",
) -> Callable[[Callable[..., LoggedEventReturnValue[T]]], Callable[..., T]]:
    """
    A decorator that ensures each call is logged in the Otter log with type ``event_type``.

    All methods decorated with this function's return value should return either ``None`` or an
    instance of the :py:class:`LoggedEventReturnValue`` dataclass.
    """

    @wrapt.decorator
    def event_logger(
        wrapped: Callable[..., Optional[LoggedEventReturnValue[T]]],
        self: "Notebook",
        args: tuple[Any],
        kwargs: dict[str, Any],
    ) -> T:
        """
        Runs a method, catching any errors and logging the call. Returns the unwrapped return value
        of the wrapped function.
        """
        try:
            ret: Optional[LoggedEventReturnValue[T]] = wrapped(*args, **kwargs)

        except Exception as e:
            self._log_event(event_type, success=False, error=e)
            raise e

        if ret is None:
            ret = LoggedEventReturnValue(None)
        if not isinstance(ret, LoggedEventReturnValue):
            raise TypeError(
                f"Method decorated by logs_event returned value of invalid type: {type(ret)}"
            )

        else:
            self._log_event(
                event_type,
                results=ret.return_value,
                question=ret.question,
                shelve_env=ret.shelve_env,
            )

        return ret.return_value

    return event_logger


def list_test_files(tests_dir: str) -> list[str]:
    """
    Find all of the test files in the specified directory (that is, all ``.py`` files that are not
    named ``__init__.py``) and return their paths in a sorted list.

    Args:
        tests_dir (``str``): the path to the tests directory

    Returns:
        ``list[str]``: the sorted list of all test file paths in ``tests_dir``
    """
    return sorted([file for file in glob(os.path.join(tests_dir, "*.py")) if file != "__init__.py"])


def list_available_tests(tests_dir: str, nbmeta_config: NBMetadataConfig) -> list[str]:
    """
    Get a list of available questions by searching the tests directory (if present) or the notebook
    metadata.

    Args:
        tests_dir (``str``): the path to the tests directory
        nbmeta_config (``otter.nbmeta_config.NBMetadataConfig``): the notebook metadata config

    Returns:
        ``list[str]``: the sorted list of question names
    """
    get_stem = lambda p: os.path.splitext(os.path.basename(p))[0]

    if tests_dir and os.path.isdir(tests_dir):
        tests = map(get_stem, list_test_files(tests_dir))

    else:
        tests = list(nbmeta_config.tests.keys())

    return sorted(tests)


def resolve_test_info(
    tests_dir: str,
    nb_path: Optional[str],
    tests_url_prefix: Optional[str],
    question: str,
) -> tuple[str, Optional[str]]:
    """
    Determine the test path and test name.

    If ``tests_url_prefix`` is specified, the test file is downloaded from the URL
    ``{tests_url_prefix}/{question}.py`` and saved to the file ``{tests_dir}/{question}.py``. If
    ``tests_dir`` does not already exist, it is created.

    Args:
        tests_dir (``str``): the path to the directory of tests
        nb_path (``str | None``): the path to the notebook
        tests_url_prefix (``str | None``): the prefix of a URL to the test file
        question (``str``): the question name

    Returns:
        ``tuple[str, str | None]``: the test path and test name (if applicable)
    """
    if tests_url_prefix is not None:
        if not IPythonInterpreter.PYOLITE.value.running():
            raise ValueError("Downloading test files from URLs is only supported on JupyterLite")

        import pyodide

        test_url = (
            f"{tests_url_prefix}{'/' if not tests_url_prefix.endswith('/') else ''}{question}.py"
        )
        text = pyodide.open_url(test_url).getvalue()

        os.makedirs(tests_dir, exist_ok=True)

        test_path = os.path.join(tests_dir, f"{question}.py")
        test_name = None

        with open(test_path, "w+") as f:
            f.write(text)

    elif tests_dir and os.path.isdir(tests_dir):
        if not os.path.isfile(os.path.join(tests_dir, question + ".py")):
            raise FileNotFoundError(f"Test {question} does not exist")

        test_path = os.path.join(tests_dir, question + ".py")
        test_name = None

    else:
        if nb_path is None:
            raise ValueError("Tests directory does not exist and no notebook path provided")

        test_path = nb_path
        test_name = question

    return test_path, test_name


def display_pdf_confirmation_widget(
    message: Optional[str], pdf_error: Optional[Exception], callback: Callable[..., Any]
) -> None:
    """
    Display a widget to the user to acknowledge that a PDF will not be included in their submission
    zip.

    Args:
        message (``str | None``): a custom message to use
        callback (``callable``): a callback function to execute after the user ACKs
    """
    o = Output()

    def wrapped_callback(*_: tuple[Any]):
        with o:
            callback()

    if not message:
        message = (
            "Your notebook could not be exported as a PDF. To continue exporting your "
            "submission, please click the button below."
        )

    message_html = f"""<p style="margin: 0">{message}</p>"""
    if pdf_error is not None:
        message_html += f"""<pre>{format_exception(pdf_error)}</pre>"""

    t = HTML(message_html)
    b = Button(description="Continue export", button_style="warning")
    b.on_click(wrapped_callback)
    m = HTML("""<div style="height: 10px; width: 100%"></div>""")

    display(VBox([t, b, m, o]))


if TYPE_CHECKING:
    from .logs import EventType
    from .notebook import Notebook



"""Exception-based test files"""

import inspect
import pathlib

from functools import lru_cache
from textwrap import indent
from types import CodeType
from typing import Any, Callable, Optional, Union

from .abstract_test import TestCase, TestCaseResult, TestFile


class test_case:
    """
    A test case function decorator for Otter's exception-based test file format.

    Holds metadata for the test cases, as well as the test case function itself, and handles calling
    the function.
    """

    name: Optional[str]
    """the name of test case"""

    points: Optional[Union[int, float]]
    """the point value of the test case, if applicable"""

    hidden: bool
    """whether the test case is hidden"""

    success_message: Optional[str]
    """a message to display to students if the test case passes"""

    failure_message: Optional[str]
    """a message to display to students if the test case fails"""

    test_func: Callable[..., None]
    """the test case function being decorated"""

    def __init__(
        self,
        name: Optional[str] = None,
        points: Optional[Union[int, float]] = None,
        hidden: bool = False,
        success_message: Optional[str] = None,
        failure_message: Optional[str] = None,
    ):
        self.name = name
        self.points = points
        self.hidden = hidden
        self.success_message = success_message
        self.failure_message = failure_message
        self.test_func = lambda: None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        return (
            self.name == other.name
            and self.points == other.points
            and self.hidden == other.hidden
            and self.success_message == other.success_message
            and self.failure_message == other.failure_message
        )

    def __call__(self, test_func: Callable[..., None]) -> "test_case":
        """
        Wrap a test case function as a decorator.
        """
        self.test_func = test_func
        return self

    def to_dataclass(self) -> TestCase:
        """
        Convert this test case to a ``TestCase`` named tuple for use in Otter's test file internals.

        Returns:
            ``otter.test_files.abstract_test.TestCase``: the test case named tuple
        """
        return TestCase(
            name=self.name,
            body=self,
            hidden=self.hidden,
            points=self.points,
            success_message=self.success_message,
            failure_message=self.failure_message,
        )

    def _get_func_params(self) -> list[str]:
        """
        Get the list of parameters expected by the decorated test case function.

        Returns:
            ``list[str]``: the function argument names
        """
        return list(inspect.signature(self.test_func).parameters.keys())

    def call_func(self, global_environment: dict[str, Any]):
        """
        Call the underlying test case function, passing in parameters from the global environment.

        If the signature of ``self.test_func`` contains a parameter called ``env``, the environment
        is passed in. For all other parameters, that value from the global environment is passed in,
        defaulting to ``None`` if the key is not present. Thus, a function with the signature

        .. code-block:: python

            foo(env, func_a, func_b, obj)

        would get called as

        .. code-block:: python

            foo(
                env = global_environment,
                func_a = global_environment.get("func_a"),
                func_b = global_environment.get("func_b"),
                obj = global_environment.get("obj"),
            )

        Args:
            global_environment (``dict[str, object]``): the global environment from which to
                retrieve values for the ``test_func`` arguments
        """
        args = self._get_func_params()
        call_kwargs = {
            arg: (global_environment if arg == "env" else global_environment.get(arg, None))
            for arg in args
        }
        self.test_func(**call_kwargs)

    def __getstate__(self) -> dict[str, Any]:
        """
        Creates a representation of the state of the instance. The attributes of the object are
        packaged into a dictionary representation excluding certain attributes. ``test_func`` is
        excluded because it cannot be pickled.

        Returns:
            ``dict``: a dictionary representation of the instance's state
        """
        state = self.__dict__.copy()
        if "test_func" in state:
            del state["test_func"]
        return state


class ExceptionTestFile(TestFile):
    """
    A single exception-based test file for Otter.
    """

    source: str = ""
    """the test file contents"""

    @property
    @lru_cache(1)
    def source_lines(self):
        """
        the lines of ``self.source``
        """
        return self.source.split("\n")

    def _generate_error_message(self, excp: Exception, context: int = 0) -> str:
        """
        Generate an error message including the line that errored from ``self.source``.

        Args:
            excp (``Exception``): the exception to generate the message for
            context (``int``): a number of extra lines of context to include from above
                and below the line that caused the exception

        Returns:
            ``str``: the error message
        """
        line_idx = excp.__traceback__.tb_next.tb_next.tb_lineno - 1
        l, h = max(0, line_idx - context), min(len(self.source_lines), line_idx + context + 1)
        lines = indent("\n".join(self.source_lines[l:h]), "  ")
        err_msg = str(excp).strip("'")
        return f"Error at line {line_idx + 1} in test {self.name}:\n{lines}\n{type(excp).__name__}: {err_msg}"

    def run(self, global_environment: dict[str, Any]):
        """
        Run the test cases against ``global_environment``, saving the results in
        ``self.test_case_results``.

        Arguments:
            global_environment (``dict[str, Any]``): result of executing a Python notebook/script
        """
        self.test_case_results = []
        for tc in self.test_cases:
            test_case = tc.body
            passed, message = True, "✅ Test case passed"
            try:
                test_case.call_func(global_environment)
            except Exception as e:
                passed, message = False, "❌ Test case failed\n" + self._generate_error_message(e)

            self.test_case_results.append(
                TestCaseResult(test_case=tc, message=message, passed=passed)
            )

    @staticmethod
    def _compile_string(s: str, path: str = "<string>") -> CodeType:
        """
        Compile a string for execution.

        Args:
            s (``str``): the string to compile
            path (``str``): the path to the test file

        Returns:
            ``code``: the compiled code of the file
        """
        return compile(s, path, "exec")

    @classmethod
    def _from_compiled_code(cls, code: CodeType, path: str = "") -> "ExceptionTestFile":
        """
        Parse a compiled exception-based test file and return an ``ExceptionTestFile``.

        Args:
            code (``code``): the compiled code of the test file
            path (``str``): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        env = {}
        exec(code, env)

        if "name" not in env:
            raise ValueError(f"Test file {path} does not define 'name'")

        name = env["name"]
        points = env.get("points", None)
        all_or_nothing = env.get("all_or_nothing", False)
        test_cases = []
        for _, v in env.items():
            if isinstance(v, test_case):
                if v.name is None:
                    v.name = f"{name} - {len(test_cases) + 1}"
                tc = v.to_dataclass()
                test_cases.append(tc)

        test_cases = cls.resolve_test_file_points(points, test_cases)

        path = str(pathlib.Path(path).as_posix())
        return cls(name, path, test_cases, all_or_nothing=all_or_nothing)

    @classmethod
    def from_string(cls, s: str, path: str = "<string>") -> "ExceptionTestFile":
        """
        Parse an exception-based test file as a string and return an ``ExceptionTestFile``.

        Args:
            s (``str``): the test file contents
            path (``str``): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        code = cls._compile_string(s, path=path)
        instc = cls._from_compiled_code(code, path=path)
        instc.source = s
        return instc

    @classmethod
    def from_file(cls, path: str) -> "ExceptionTestFile":
        """
        Parse an exception-based test file and return an ``ExceptionTestFile``.

        Args:
            path (``str``): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        with open(path) as f:
            source = f.read()
        return cls.from_string(source, path=path)

    @classmethod
    def from_metadata(cls, s: str, path: str) -> "ExceptionTestFile":
        """
        Parse an exception-based test file from its data stored in a notebook's metadata.

        Args:
            s (``str``): the test file contents from the notebook metadata
            path (``str``): the path to the notebook

        Returns:
            ``ExceptionTestFile``: the parsed test file.
        """
        return cls.from_string(s, path=path)



"""A class to collect test file data from R autograding with Ottr"""

from typing import Any

from .abstract_test import TestFile


class OttrTestFile(TestFile):

    def run(self, global_environment: dict[str, Any]):
        raise NotImplementedError("Ottr test files cannot be run from Python")

    @classmethod
    def from_file(cls, path: str) -> "OttrTestFile":
        raise NotImplementedError("Cannot create Ottr test files from a file in Python")

    @classmethod
    def from_metadata(cls, s: Any, path: str) -> "TestFile":
        raise NotImplementedError("Cannot create Ottr test files from a file in Python")



"""Execution and grading internals for Otter-Grader"""

import nbformat
import os
import pickle
import tempfile

from traitlets.config import Config
from typing import Optional, TYPE_CHECKING

from .checker import Checker
from .logging import start_server
from ..plugins import PluginCollection
from ..test_files import GradingResults
from ..utils import NBFORMAT_VERSION


__all__ = ["Checker", "grade_notebook"]


def grade_notebook(
    submission_path: str,
    *,
    tests_glob: Optional[list[str]] = None,
    ignore_errors: bool = True,
    script: bool = False,
    cwd: Optional[str] = None,
    test_dir: Optional[str] = None,
    seed: Optional[int] = None,
    seed_variable: Optional[str] = None,
    log: Optional["Log"] = None,
    variables: Optional[dict[str, str]] = None,
    plugin_collection: Optional[PluginCollection] = None,
    force_python3_kernel: bool = True,
):
    """
    Grade an assignment file and return grade information.

    Args:
        submission_path (``str``): path to a single notebook or Python script
        tests_glob (``list[str] | NOne``): paths of test files that should be run; tests that are
            included in this list that have not already been run by the submission are run after
            executing the submission
        ignore_errors (``bool``): whether errors in execution should be ignored
        script (``bool``): whether the ``submission_path`` is a Python script
        cwd (``str | None``): working directory of execution to be appended to ``sys.path`` before
            executing the submission
        test_dir (``str | None``): path to directory of tests in the grading environment
        seed (``int | None``): random seed for intercell seeding
        seed_variable (``str | None``): a variable name to override with the seed
        log (``otter.check.logs.Log``): an Otter log from which to deserialize environments to grade
            questions
        variables (``dict[str, str] | None``): map of variable names -> type string to use for type
            checking values deserialized from ``log``
        plugin_collection (``otter.plugins.PluginCollection | None``): a set of plugins to run the
            ``before_execution`` and ``after_grading`` events on this submission

    Returns:
        ``otter.test_files.GradingResults``: the results of grading
    """
    from nbconvert.preprocessors import ExecutePreprocessor

    from .preprocessor import GradingPreprocessor

    if tests_glob is None:
        tests_glob = []

    if not script:
        nb = nbformat.read(submission_path, as_version=NBFORMAT_VERSION)

    else:
        with open(submission_path) as f:
            nb = f.read()

        nb = nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell(nb)])

    if plugin_collection is not None:
        nb = plugin_collection.before_execution(nb)

    results_handle, results_file = tempfile.mkstemp(suffix=".pkl")

    try:
        c = Config()

        (host, port), stop_server = start_server()

        try:
            # GradingPreprocessor config
            c.GradingPreprocessor.cwd = cwd
            c.GradingPreprocessor.test_dir = test_dir
            c.GradingPreprocessor.tests_glob = tests_glob
            c.GradingPreprocessor.results_path = results_file
            c.GradingPreprocessor.seed = seed
            c.GradingPreprocessor.seed_variable = seed_variable
            c.GradingPreprocessor.otter_log = log
            c.GradingPreprocessor.variables = variables
            c.GradingPreprocessor.logging_server_host = host
            c.GradingPreprocessor.logging_server_port = port
            c.GradingPreprocessor.force_python3_kernel = force_python3_kernel

            # ExecutePreprocessor config
            c.ExecutePreprocessor.allow_errors = ignore_errors

            gp = GradingPreprocessor(config=c)
            ep = ExecutePreprocessor(config=c)

            nb, _ = gp.preprocess(nb)
            executed_nb, _ = ep.preprocess(nb)

        finally:
            stop_server()
            gp.cleanup()

        os.close(results_handle)

        try:
            with open(results_file, "rb") as f:
                results = pickle.load(f)
        except Exception as e:
            results = GradingResults.without_results(e)

        if not isinstance(results, GradingResults):
            raise TypeError(
                "Results deserialized from grading notebook were not a GradingResults instance"
            )

        results.notebook = executed_nb

        if plugin_collection is not None:
            plugin_collection.run("after_grading", results)

        return results

    finally:
        os.remove(results_file)


if TYPE_CHECKING:
    from ..check.logs import Log



"""Plugin replacement for Otter Assign"""

import copy
import nbformat as nbf
import yaml

from ..utils import get_source


BEGIN = "# BEGIN PLUGIN"
END = "# END PLUGIN"
BEGIN_EXPORT = "# BEGIN PLUGIN EXPORT"


def replace_plugins(lines: list[str]) -> list[str]:
    """
    Replace plugin blocks with plugin calls in a cell's source (a list of strings).

    Args:
        lines (``list[str]``): the cell source

    Returns:
        ``list[str]``: a copy of ``lines`` with plugin calls
    """
    starts, ends = [], []
    stripped = [[]]
    exports = []
    plugin = False
    for i, line in enumerate(lines):
        if line.rstrip().endswith(END):
            assert plugin, f"END PLUGIN without BEGIN PLUGIN found in {lines}"
            plugin = False
            ends.append(i)
            stripped.append([])

        elif line.rstrip().endswith(BEGIN):
            assert not plugin, f"Nested plugins found in {lines}"
            starts.append(i)
            exports.append(False)
            plugin = True

        elif line.rstrip().endswith(BEGIN_EXPORT):
            assert not plugin, f"Nested plugins found in {lines}"
            starts.append(i)
            exports.append(True)
            plugin = True

        elif plugin:
            stripped[len(starts) - 1].append(line)

    assert (
        len(stripped) == len(starts) + 1 == len(ends) + 1 == len(exports) + 1
    ), f"Error processing plugins in {lines}"
    assert all(s < e for s, e in zip(starts, ends))

    starts.reverse()
    ends.reverse()
    stripped.reverse()
    stripped = stripped[1:]

    lines = lines.copy()

    for i, (s, e) in enumerate(zip(starts, ends)):
        config = yaml.full_load("\n".join(stripped[i]))
        export = exports[i]
        pg = config["plugin"]
        args = ", ".join(config.get("args", []))
        kwargs = ", ".join([f"{k}={v}" for k, v in config.get("kwargs", {}).items()])

        call = ("run_plugin", "add_plugin_files")[export]

        call = f'grader.{call}("{pg}"'
        if args:
            call += f", {args}"
        if kwargs:
            call += f", {kwargs}"
        call += ")"

        del lines[s : e + 1]
        lines.insert(s, call)

    return lines


def replace_plugins_with_calls(nb: nbf.NotebookNode) -> nbf.NotebookNode:
    """
    Replace all plugin blocks in a notebook with plugin calls.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``nbformat.NotebookNode``: a copy of ``nb`` with plugin blocks replaced
    """
    nb = copy.deepcopy(nb)

    for cell in nb["cells"]:
        cell["source"] = "\n".join(replace_plugins(get_source(cell)))

    return nb



"""Support for OK-formatted test files"""

import doctest
import io
import os
import pathlib

from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent
from typing import Any

from .abstract_test import TestCase, TestCaseResult, TestFile
from ..utils import hide_outputs


def run_doctest(
    name: str, doctest_string: str, global_environment: dict[str, Any]
) -> tuple[bool, str]:
    """
    Run a single test with given ``global_environment``. Returns ``(True, '')`` if the doctest passes.
    Returns ``(False, failure_message)`` if the doctest fails.

    Args:
        name (``str``): name of doctest
        doctest_string (``str``): doctest in string form
        global_environment (``dict[str, Any]``): global environment resulting from the execution of
            a python script/notebook

    Returns:
        ``tuple[bool, str]``: results from running the test
    """
    examples = doctest.DocTestParser().parse(doctest_string, name)
    test = doctest.DocTest(
        [e for e in examples if isinstance(e, doctest.Example)],
        global_environment,
        name,
        None,
        None,
        doctest_string,
    )

    doctestrunner = doctest.DocTestRunner(verbose=True)

    run_results = io.StringIO()
    with redirect_stdout(run_results), redirect_stderr(run_results), hide_outputs():
        doctestrunner.run(test, clear_globs=False)
    with open(os.devnull, "w") as f, redirect_stderr(f), redirect_stdout(f):
        result = doctestrunner.summarize(verbose=True)
    # An individual test can only pass or fail
    if result.failed == 0:
        return (True, "")
    else:
        return False, run_results.getvalue()


class OKTestFile(TestFile):
    """
    A single OK-formatted test file for Otter.
    """

    def run(self, global_environment: dict[str, Any]):
        """
        Run the test cases on ``global_environment``, saving the results in
        ``self.test_case_results``.

        Arguments:
            ``global_environment`` (``dict``): result of executing a Python notebook/script
        """
        self.test_case_results = []
        for i, test_case in enumerate(self.test_cases):
            passed, result = run_doctest(
                self.name + " " + str(i), test_case.body, global_environment
            )
            if passed:
                result = "✅ Test case passed"
            else:
                result = "❌ Test case failed\n" + result

            self.test_case_results.append(
                TestCaseResult(
                    test_case=test_case,
                    message=result,
                    passed=passed,
                )
            )

    @classmethod
    def from_spec(cls, test_spec: dict[str, Any], path: str = "") -> "OKTestFile":
        """
        Parse an OK-formatted ``dict`` and return an ``OKTestFile``.

        Args:
            test_spec (``dict[str, Any]``): the OK-formatted ``dict``
            path (``str``): the path to the test file this ``dict`` was parsed from

        Returns:
            ``OKTestFile``: the new ``OKTestFile`` object created from the given file
        """
        # Make sure there is a name
        assert "name" in test_spec

        # Do not support multiple suites in the same file
        assert len(test_spec["suites"]) == 1

        test_suite = test_spec["suites"][0]

        # Only support doctest. I am unsure if other tests are implemented
        assert test_suite.get("type", "doctest") == "doctest"

        # Not setup and teardown supported
        assert not bool(test_suite.get("setup"))
        assert not bool(test_suite.get("teardown"))

        test_cases = []
        for i, test_case in enumerate(test_spec["suites"][0]["cases"]):
            test_cases.append(
                TestCase(
                    name=test_case.get("name", f"{test_spec['name']} - {i + 1}"),
                    body=dedent(test_case["code"]),
                    hidden=test_case.get("hidden", True),
                    points=test_case.get("points", None),
                    success_message=test_case.get("success_message", None),
                    failure_message=test_case.get("failure_message", None),
                )
            )

        # resolve point values for each test case
        spec_pts = test_spec.get("points", None)
        test_cases = cls.resolve_test_file_points(spec_pts, test_cases)

        # convert path into PurePosixPath for test name
        path = str(pathlib.Path(path).as_posix())

        # grab whether the tests are all-or-nothing
        all_or_nothing = test_spec.get("all_or_nothing", False)

        return cls(test_spec["name"], path, test_cases, all_or_nothing)

    @classmethod
    def from_metadata(cls, s: Any, path: str) -> "OKTestFile":
        """
        Parse an OK test file from its data stored in a notebook's metadata.

        Args:
            spec (``dict``): the test data from the notebook metadata
            path (``str``): the path to the notebook

        Returns:
            ``OKTestFile``: the parsed test file.
        """
        return cls.from_spec(s, path=path)

    @classmethod
    def from_file(cls, path: str) -> "OKTestFile":
        """
        Parse an OK-formatted test file and return an ``OKTestFile``.

        Args:
            path (``str``): the path to the test file

        Returns:
            ``OKTestFile``: the new ``OKTestFile`` object created from the given file
        """
        # ok test files are python files, with a global 'test' defined
        test_globals = {}
        with open(path) as f:
            exec(f.read(), test_globals)

        test_spec = test_globals["test"]

        return cls.from_spec(test_spec, path=path)



"""A logging server for receiving logs from the grading process"""

import logging
import logging.handlers
import pickle
import socket
import socketserver
import struct
import threading

from ..logging import get_level, get_logger


LOGGER = get_logger(__name__)


class LogLevelFilter(logging.Filter):
    """
    A simple ``logging.Filter`` that checks that the level of the provided ``LogRecord`` is at least
    as high as the configured log level.
    """

    def filter(self, record: logging.LogRecord):
        return record.levelno >= get_level()


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """
    A handler for a streaming logging request.

    Copied from https://docs.python.org/3/howto/logging-cookbook.html#sending-and-receiving-logging-events-across-a-network.
    """

    filter = LogLevelFilter()

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unpickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handle_log_record(record)

    def unpickle(self, data: bytes):
        return pickle.loads(data)

    def handle_log_record(self, record: logging.LogRecord):
        logger = get_logger(record.name)
        if self.filter.filter(record):
            logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver.

    Copied from https://docs.python.org/3/howto/logging-cookbook.html#sending-and-receiving-logging-events-across-a-network.
    """

    allow_reuse_address = True
    abort = False

    def __init__(
        self,
        host: str = "localhost",
        port: int = 0,
        handler: type[logging.Handler] = LogRecordStreamHandler,
    ):
        super().__init__((host, port), handler)

    def serve_until_stopped(self):
        import select

        abort = False
        while not abort:
            rd = None
            try:
                rd, _, _ = select.select([self.socket.fileno()], [], [], 1)
            except socket.error:
                pass
            if rd:
                self.handle_request()
            abort = self.abort


def start_server():
    """
    Start a TCP socket for receiving logs from the notebook being executed. Returns a tuple of
    ``((server host, server port), stop server callback)``.

    Returns:
        ``tuple[tuple[str, int], callable]``: a tuple containing a tuple with the server host and
            port as its first element and a callback to stop the server as its second
    """
    tcpserver = LogRecordSocketReceiver()
    host, port = tcpserver.server_address
    LOGGER.debug(f"Starting execution logging TCP server in background at {host}:{port}")
    t = threading.Thread(target=tcpserver.serve_until_stopped)
    t.start()

    def stop_server():
        LOGGER.debug("Stopping execution logging TCP server")
        tcpserver.abort = True

    return tcpserver.server_address, stop_server



"""Assignment creation tool for Otter-Grader"""

import os
import pathlib
import warnings

from typing import Optional

from .assignment import Assignment
from .output import write_output_directories
from .utils import run_generate_autograder, run_tests, write_otter_config_file
from .. import logging
from ..export import export_notebook
from ..plugins import PluginCollection
from ..utils import chdir, get_relpath, knit_rmd_file


__all__ = ["Assignment", "main"]


LOGGER = logging.get_logger(__name__)


def main(
    master: str,
    result: str,
    *,
    no_pdfs: bool = False,
    no_run_tests: bool = False,
    username: Optional[str] = None,
    password: Optional[str] = None,
    debug: Optional[str] = False,
):
    """
    Runs Otter Assign on a master notebook.

    Args:
        master (``str``): path to master notebook
        result (``str``): path to result directory
        no_pdfs (``bool``): whether to ignore any configurations indicating PDF generation for this run
        no_run_tests (``bool``): prevents Otter tests from being automatically run on the solutions
            notebook
        username (``str | None``): a username for Gradescope for generating a token
        password (``str | None``): a password for Gradescope for generating a token
        debug (``bool``): whether to run in debug mode (without ignoring errors during testing)
    """
    LOGGER.debug(f"User-specified master path: {master}")
    LOGGER.debug(f"User-specified result path: {result}")
    masterp, resultp = pathlib.Path(os.path.abspath(master)), pathlib.Path(os.path.abspath(result))

    assignment = Assignment(require_valid_keys=True)

    resultp = get_relpath(masterp.parent, resultp)

    assignment.master, assignment.result = masterp, resultp
    LOGGER.debug(f"Normalized master path: {master}")
    LOGGER.debug(f"Normalized result path: {result}")

    with chdir(assignment.master.parent):
        LOGGER.info("Generating views")
        write_output_directories(assignment)

        # update seed variables
        if assignment.seed.variable:
            LOGGER.debug("Processing seed dict")
            assignment.generate.seed = assignment.seed.autograder_value
            assignment.generate.seed_variable = assignment.seed.variable
            LOGGER.debug("Added seed information to assignment.generate")

        # check that we have a seed if needed
        if assignment.seed_required:
            LOGGER.debug("Assignment seed is required")
            if not isinstance(assignment.generate.seed, int):
                raise RuntimeError("Seeding cell found but no or invalid seed provided")

        plugins, pc = assignment.plugins, None
        if plugins:
            LOGGER.debug("Processing plugins")
            pc = PluginCollection(plugins, "", {})
            pc.run("during_assign", assignment)

            LOGGER.debug("Adding plugin configurations to Otter Generate configuration")
            assignment.generate.plugins = assignment.generate.plugins + plugins

        LOGGER.info("Generating autograder zipfile")
        run_generate_autograder(assignment, username, password, plugin_collection=pc)

        # generate PDF of solutions
        if assignment.solutions_pdf and not no_pdfs:
            LOGGER.info("Generating solutions PDF")
            filtering = assignment.solutions_pdf == "filtered"

            src = os.path.abspath(str(assignment.get_ag_path(assignment.master.name)))
            dst = os.path.abspath(str(assignment.get_ag_path(assignment.master.stem + "-sol.pdf")))

            if not assignment.is_rmd:
                LOGGER.debug(f"Exporting {src} as notebook to {dst}")
                LOGGER.debug("Attempting PDF via HTML export")
                export_notebook(
                    src,
                    dest=dst,
                    filtering=filtering,
                    pagebreaks=filtering,
                    exporter_type="html",
                )
                LOGGER.debug("PDF via HTML export successful")

            else:
                LOGGER.debug(f"Knitting {src} to {dst}")
                if filtering:
                    raise ValueError("Filtering is not supported with RMarkdown assignments")

                knit_rmd_file(src, dst)

        # generate a tempalte PDF for Gradescope
        if assignment.template_pdf and not no_pdfs:
            LOGGER.info("Generating template PDF")

            src = os.path.abspath(str(assignment.get_ag_path(assignment.master.name)))
            dst = os.path.abspath(
                str(assignment.get_ag_path(assignment.master.stem + "-template.pdf"))
            )

            if not assignment.is_rmd:
                LOGGER.debug("Attempting PDF via LaTeX export")
                export_notebook(
                    src,
                    dest=dst,
                    filtering=True,
                    pagebreaks=True,
                    exporter_type="latex",
                )
                LOGGER.debug("PDF via LaTeX export successful")

            else:
                raise ValueError(
                    f"Filtering is not supported with RMarkdown assignments; use "
                    + "solutions_pdf to generate a Gradescope template instead."
                )

        # generate the .otter file if needed
        if not assignment.is_rmd and assignment.save_environment:
            LOGGER.debug("Processing environment serialization configuration")
            if assignment.is_r:
                warnings.warn(
                    "Otter Service and serialized environments are unsupported with R, "
                    "configurations ignored"
                )
            else:
                write_otter_config_file(assignment)

        # run tests on autograder notebook
        if assignment.run_tests and not no_run_tests:
            LOGGER.info("Running tests against the solutions notebook")

            run_tests(assignment, debug=debug)

            LOGGER.info("All autograder tests passed.")

        else:
            LOGGER.info("Skipping tests")



"""IPython notebook API for Otter Check"""

import datetime as dt
import inspect
import json
import nbformat as nbf
import os
import sys
import warnings
import zipfile

from glob import glob
from IPython.display import display, HTML
from textwrap import indent
from typing import Any, ClassVar, Optional, Union

from .logs import EventType, Log, LogEntry
from .utils import (
    display_pdf_confirmation_widget,
    grade_zip_file,
    grading_mode_disabled,
    incompatible_with,
    IPythonInterpreter,
    list_available_tests,
    LoggedEventReturnValue,
    logs_event,
    resolve_test_info,
    save_notebook,
)
from ..execute import Checker
from ..export import export_notebook
from ..logging import Loggable
from ..nbmeta_config import NBMetadataConfig
from ..plugins import PluginCollection
from ..test_files import GradingResults, TestFile


OTTER_LOG_FILENAME = ".OTTER_LOG"
_ZIP_NAME_FILENAME = "__zip_filename__"


class Notebook(Loggable):
    """
    Notebook class for in-notebook autograding

    Args:
        nb_path (``str | None``): path to the notebook being run
        tests_dir (``str``): path to tests directory
        tests_url_prefix (``str | None``): a URL prefix to use to download test files
    """

    _grading_mode: ClassVar[bool] = False
    """whether Otter is currently in grading mode"""

    _shelve: ClassVar[bool] = False
    """whether to shelve the global environment in log entries"""

    _tests_dir_override: ClassVar[Optional[str]] = None
    """an override for the path to the tests directory"""

    _notebook: Optional[str]
    """the path to the notebook file"""

    _tests_dir: str
    """the path to the tests directory"""

    _tests_url_prefix: Optional[str]
    """a URL prefix to use for downloading test files"""

    _addl_files: list[str]
    """a list of additional file paths to include in the exported submission zip"""

    _plugin_collections: dict[str, PluginCollection]
    """a cache of ``PluginCollection`` objects for each plugin invoked in the notebook"""

    interpreter: Optional[IPythonInterpreter]
    """the interpreter that is currently running (if not the standard IPython interpreter)"""

    _nbmeta_config: NBMetadataConfig
    """the metadata config from the notebook"""

    _config: Optional[dict[str, Any]] = None
    """the config loaded from the .otter file if there was one"""

    _ignore_modules: Optional[list[str]] = None
    """a list of modules to ignore when serializing environments"""

    _vars_to_store: Optional[dict[str, str]] = None
    """a map of var names -> type name to use when serializing environments"""

    @logs_event(EventType.INIT)
    def __init__(
        self,
        nb_path: Optional[str] = None,
        tests_dir: str = "./tests",
        tests_url_prefix: Optional[str] = None,
    ):
        interpreter = None
        for i in IPythonInterpreter:
            if i.value.running():
                interpreter = i
                break

        self.interpreter = interpreter

        if self.interpreter is IPythonInterpreter.COLAB and not os.path.isdir(tests_dir):
            raise ValueError(f"Tests directory {tests_dir} does not exist")

        cls = type(self)
        if cls._tests_dir_override is not None:
            self._tests_dir = cls._tests_dir_override
        else:
            self._tests_dir = tests_dir

        self._notebook = nb_path
        self._notebook = self._resolve_nb_path(nb_path, fail_silently=True)
        self._tests_url_prefix = tests_url_prefix
        self._addl_files = []
        self._plugin_collections = {}

        # assume using otter service if there is a .otter file
        otter_configs = glob("*.otter")
        if otter_configs:
            # check that there is only 1 config
            assert len(otter_configs) == 1, "More than 1 otter config file found"

            # load in config file
            with open(otter_configs[0], encoding="utf-8") as f:
                self._config = json.load(f)

            if not isinstance(self._config, dict):
                raise ValueError(f'"{otter_configs[0]}" does not contain a valid config')

            type(self)._shelve = self._config.get("save_environment", False)
            self._ignore_modules = self._config.get("ignore_modules", [])
            self._vars_to_store = self._config.get("variables", None)

            self._notebook = self._config["notebook"]

        if self._notebook:
            self._nbmeta_config = NBMetadataConfig.from_notebook(
                nbf.read(self._notebook, nbf.NO_CONVERT)
            )
        else:
            self._nbmeta_config = NBMetadataConfig()

    @classmethod
    def init_grading_mode(cls, tests_dir: str):
        logger = cls._get_logger()
        logger.info("Entering Notebook grading mode")
        logger.debug(f"Overriding tests directory: {tests_dir}")
        cls._grading_mode = True
        cls._tests_dir_override = tests_dir
        Checker.clear_results()
        Checker.enable_tracking()

    @incompatible_with(IPythonInterpreter.PYOLITE, throw_error=False)
    def _log_event(
        self,
        event_type: EventType,
        results: Optional[Union[GradingResults, TestFile]] = None,
        question: Optional[str] = None,
        success: bool = True,
        error: Optional[Exception] = None,
        shelve_env: Optional[dict[str, Any]] = None,
    ):
        """
        Log an event to Otter's client log file.

        Args:
            event_type (``otter.logs.EventType``): the type of event
            results (``otter.test_files.TestFile | None``): the results of any checks recorded
                by the entry
            question (``str | None``): the question name for this check
            success (``bool``): whether the operation was successful
            error (``Exception | None``): the exception thrown by the operation, if applicable
            shelve_env (``dict[str, Any] | None``): the environment to shelve
        """
        entry = LogEntry(
            event_type, results=results, question=question, success=success, error=error
        )

        if type(self)._shelve and event_type == EventType.CHECK:
            if not isinstance(shelve_env, dict):
                raise TypeError(f"shelve_env has an invalid type: {type(shelve_env)}")

            entry.shelve(
                shelve_env,
                delete=True,
                filename=OTTER_LOG_FILENAME,
                ignore_modules=self._ignore_modules,
                variables=self._vars_to_store,
            )

        entry.flush_to_file(OTTER_LOG_FILENAME)

    def _resolve_nb_path(
        self, nb_path: Optional[str], fail_silently: bool = False
    ) -> Optional[str]:
        """
        Attempts to resolve the path to the notebook being run. If ``nb_path`` is ``None``,
        ``self._notebook`` is checked, then the working directory is searched for ``.ipynb`` files.

        Args:
            nb_path (``str | None``): path to the notebook
            fail_silently (``bool``): if true, the method does not fail the notebook path can't be
                resolved

        Returns:
            ``str | None``: resolved notebook path or ``None`` if it can't be resolved and
                ``fail_silently`` was true

        Raises:
            ``ValueError``: if no notebooks or too many notebooks are found and ``fail_silently`` is
                false
        """
        # if in grading mode, don't attempt to resolve the notebook path, since the tests path was
        # already overridden in __init__
        if type(self)._grading_mode:
            return nb_path

        if nb_path is None and self._notebook is not None:
            nb_path = self._notebook
            if not os.path.isfile(nb_path):
                raise ValueError(
                    f"Expected a notebook file named '{nb_path}' but no such file found"
                )

        elif nb_path is None and glob("*.ipynb"):
            notebooks = glob("*.ipynb")
            if len(notebooks) == 1:
                nb_path = notebooks[0]

        if nb_path is None and not fail_silently:
            raise ValueError("Could not resolve notebook path")

        return nb_path

    @logs_event(EventType.CHECK)
    def check(
        self, question: str, global_env: Optional[dict[str, Any]] = None
    ) -> LoggedEventReturnValue[TestFile]:
        """
        Runs tests for a specific question against a global environment. If no global environment
        is provided, the test is run against the calling frame's environment.

        Args:
            question (``str``): name of question being graded
            global_env (``dict[str, object]  | None``): a global environment in which to run the
                tests

        Returns:
            ``otter.test_files.abstract_test.TestFile``: the grade for the question
        """
        self._logger.info(f"Running check for question: {question}")
        test_path, test_name = resolve_test_info(
            self._tests_dir,
            self._notebook,
            self._tests_url_prefix,
            question,
        )

        self._logger.debug(f"Resolved test path: {test_path}")
        self._logger.debug(f"Resolved test name: {test_name}")

        # raise an error for a metadata test on Colab
        if test_name is not None and self.interpreter is IPythonInterpreter.COLAB:
            raise ValueError(f"Test {question} does not exist")

        # ensure that desired test exists
        if not os.path.isfile(test_path):
            raise FileNotFoundError(f"Test {question} does not exist")

        # pass the correct global environment
        if global_env is None:
            self._logger.debug(f"Collecting calling global environment")
            frame = inspect.currentframe().f_back.f_back
            # I have NO IDEA why but in python 3.13 ONLY WHEN RUNNING TESTS (?!?!) there is an
            # additional frame for wrapt. I spent h o u r s trying to figure out why but I was
            # unsuccessful.
            if f"{os.path.sep}wrapt{os.path.sep}" in frame.f_code.co_filename:
                frame = frame.f_back
            global_env = frame.f_globals

        # run the check
        self._logger.debug(f"Calling checker")
        result = Checker.check(test_path, self._nbmeta_config, test_name, global_env)

        return LoggedEventReturnValue(result, question=question, shelve_env=global_env)

    @incompatible_with(IPythonInterpreter.COLAB)
    def run_plugin(
        self,
        plugin_name: str,
        *args: Any,
        nb_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Runs the plugin ``plugin_name`` with the specified arguments. Use ``nb_path`` if the path
        to the notebook is not configured.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str | None``): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin
        """
        self._logger.info(f"Running plugin {plugin_name}")
        self._logger.debug(f"Running plugin {plugin_name} with args {args} and kwargs {kwargs}")
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        pc.run("from_notebook", *args, **kwargs)

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    @logs_event(EventType.TO_PDF)
    def to_pdf(
        self,
        nb_path: Optional[str] = None,
        filtering: bool = True,
        pagebreaks: bool = True,
        display_link: bool = True,
        force_save: bool = False,
    ):
        """
        Exports a notebook to a PDF using Otter Export

        Args:
            nb_path (``str | None``): path to the notebook we want to export; will attempt to infer
                if not provided
            filtering (``bool``): set true if only exporting a subset of notebook cells to PDF
            pagebreaks (``bool``): if true, pagebreaks are included between questions
            display_link (``bool``): whether or not to display a download link
            force_save (``bool``): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
        """
        nb_path = self._resolve_nb_path(nb_path)
        self._logger.debug(f"Resolved notebook path: {nb_path}")

        if force_save:
            self._logger.debug("Attempting to force-save notebook")
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Couldn't automatically save the notebook; we recommend using File > Save & "
                    "Checkpoint and then re-running this cell. The zip file returned by this call "
                    "will use the last saved version of this notebook."
                )
            else:
                self._logger.debug("Force-save successful")

        pdf_path = export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
        self._logger.debug(f"Wrote PDF to zip file: {pdf_path}")

        if display_link:
            # create and display output HTML
            out_html = f"""
            <p>Your file has been exported. Download it by right-clicking
            <a href="{pdf_path}" target="_blank">here</a> and selecting <strong>Save Link As</strong>.
            """

            display(HTML(out_html))

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    def add_plugin_files(
        self,
        plugin_name: str,
        *args: Any,
        nb_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Runs the ``notebook_export`` event of the plugin ``plugin_name`` and tracks the file paths
        it returns to be included when calling ``Notebook.export``.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str | None``): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin
        """
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        addl_files = pc.run("notebook_export", *args, **kwargs)[0]
        if addl_files is None:
            return
        self._addl_files.extend(addl_files)

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    @logs_event(EventType.END_EXPORT)
    def export(
        self,
        nb_path: Optional[str] = None,
        export_path: Optional[str] = None,
        pdf: bool = True,
        filtering: bool = True,
        pagebreaks: bool = True,
        files: Optional[list[str]] = None,
        display_link: bool = True,
        force_save: bool = False,
        run_tests: bool = False,
        ignore_log: bool = False,
    ):
        """
        Export a submission zip file.

        Creates a submission zipfile from a notebook at ``nb_path``, optionally including a PDF
        export of the notebook and any files in ``files``.

        If ``run_tests`` is true, the zip is validated by running it against local test cases in a
        separate Python process. The results of this run are printed to stdout.

        Args:
            nb_path (``str | None``): path to the notebook we want to export; will attempt to infer
                if not provided
            export_path (``str | None``): path at which to write zipfile; defaults to
                ``{notebook name}_{timestamp}.zip``
            pdf (``bool``): whether a PDF should be included
            filtering (``bool``): whether the PDF should be filtered
            pagebreaks (``bool``): whether pagebreaks should be included between questions in the
                PDF
            files (``list[str] | None``): paths to other files to include in the zip file
            display_link (``bool``): whether or not to display a download link
            force_save (``bool``): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
            run_tests (``bool``): whether to validating the resulting submission zip against local
                test cases
            ignore_log (``bool``): whether to exclude the .OTTER_LOG file from the submission zip
        """
        self._log_event(EventType.BEGIN_EXPORT)

        files = [] if files is None else files

        nb_path = self._resolve_nb_path(nb_path)
        self._logger.debug(f"Resolved notebook path: {nb_path}")

        if force_save:
            self._logger.debug("Attempting to force-save notebook")
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Couldn't automatically save the notebook; we recommend using File > Save & "
                    "Checkpoint and then re-running this cell. The zip file returned by this call "
                    "will use the last saved version of this notebook."
                )
            else:
                self._logger.debug("Force-save successful")

        with open(nb_path, "r", encoding="utf-8") as f:
            if len(f.read().strip()) == 0:
                raise ValueError(
                    f"Notebook '{nb_path}' is empty. Please save and checkpoint your "
                    "notebook and rerun this cell."
                )

            f.seek(0)

        timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
        if export_path is None:
            zip_path = ".".join(nb_path.split(".")[:-1]) + "_" + timestamp + ".zip"
        else:
            zip_path = export_path

        self._logger.debug(f"Determined export zip path: {zip_path}")

        zf = zipfile.ZipFile(zip_path, mode="w")
        zf.write(nb_path)

        pdf_path, pdf_created, pdf_error = None, True, None
        if pdf:
            try:
                pdf_path = export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
            except Exception as e:
                pdf_error = e
            if pdf_path and os.path.isfile(pdf_path):
                pdf_created = True
                zf.write(pdf_path)
                self._logger.debug(f"Wrote PDF to zip file: {pdf_path}")
            else:
                pdf_created = False
                warnings.warn("Could not locate a PDF to include")

        def continue_export():
            if not ignore_log and os.path.isfile(OTTER_LOG_FILENAME):
                zf.write(OTTER_LOG_FILENAME)
                self._logger.debug("Added Otter log to zip file")

            zip_basename = os.path.basename(zip_path)
            zf.writestr(_ZIP_NAME_FILENAME, zip_basename)
            self._logger.debug(f"Added {_ZIP_NAME_FILENAME} to zip file: '{zip_basename}'")

            dot_otter = glob("*.otter")
            if dot_otter:
                if len(dot_otter) != 1:
                    raise ValueError("Too many .otter files (max 1 allowed)")
                dot_otter = dot_otter[0]
                zf.write(dot_otter)
                self._logger.debug(f"Added .otter file to zip file: {dot_otter}")

            for file in files:
                if os.path.isdir(file):
                    sub_files = glob(f"{file}/**/*.*", recursive=True)
                    for sub_file in sub_files:
                        zf.write(sub_file)
                else:
                    zf.write(file)
                self._logger.debug(f"Added file to zip file: {file}")

            for file in self._addl_files:
                zf.write(file)
                self._logger.debug(f"Added plugin file to zip file: {file}")

            zf.close()

            if run_tests:
                print("Running your submission against local test cases...\n")
                results = grade_zip_file(zip_path, nb_path, self._tests_dir)
                print(
                    "Your submission received the following results when run against "
                    + "available test cases:\n\n"
                    + indent(results.summary(), "    ")
                )

            if display_link:
                # create and display output HTML
                out_html = f"""
                    <p>
                        Your submission has been exported. Click
                        <a href="{zip_path}" download="{zip_path}" target="_blank">here</a> to download
                        the zip file.
                    </p>
                """

                display(HTML(out_html))

        if pdf_created or not self._nbmeta_config.require_no_pdf_confirmation:
            if pdf_error is not None:
                raise pdf_error
            continue_export()
        else:
            display_pdf_confirmation_widget(
                self._nbmeta_config.export_pdf_failure_message, pdf_error, continue_export
            )

    @grading_mode_disabled
    @logs_event(EventType.END_CHECK_ALL)
    def check_all(self):
        """
        Runs all tests on this notebook. Tests are run against the current global environment, so
        any tests with variable name collisions will fail.
        """
        self._log_event(EventType.BEGIN_CHECK_ALL)

        tests = list_available_tests(self._tests_dir, self._nbmeta_config)

        frame = inspect.currentframe().f_back.f_back
        # I have NO IDEA why but in python 3.13 ONLY WHEN RUNNING TESTS (?!?!) there is an
        # additional frame for wrapt. I spent h o u r s trying to figure out why but I was
        # unsuccessful.
        if f"{os.path.sep}wrapt{os.path.sep}" in frame.f_code.co_filename:
            frame = frame.f_back
        # do it again since there are 2 decorators on this method
        frame = frame.f_back
        if f"{os.path.sep}wrapt{os.path.sep}" in frame.f_code.co_filename:
            frame = frame.f_back

        global_env = frame.f_globals

        self._logger.debug(f"Found available tests: {', '.join(tests)}")

        results = []
        if not type(self)._shelve:
            for test_name in tests:
                results.append(self.check(test_name, global_env))

        else:
            log = Log.from_file(OTTER_LOG_FILENAME, ascending=False)
            for file in sorted(tests):
                if "__init__.py" not in file:
                    test_name = os.path.splitext(os.path.split(file)[1])[0]

                    entry = log.get_question_entry(test_name)
                    env = entry.unshelve()
                    global_env.update(env)
                    del locals()["env"]

                    result = self.check(test_name, global_env)
                    results.append((test_name, result))

        return LoggedEventReturnValue(GradingResults(results))



"""Abstract test objects for providing a schema to write and parse test cases"""

import random

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, replace
from textwrap import indent
from typing import Any, Optional, Union


@dataclass
class TestCase:
    """
    A dataclass representing a single test case in a test file.
    """

    name: str
    """the name of the test case"""

    body: Any
    """the code for the test case"""

    hidden: bool
    """whether this test case is hidden from students"""

    points: Optional[Union[int, float]]
    """the point value of this test case"""

    success_message: Optional[str]
    """a message to show to students if this test cases passes"""

    failure_message: Optional[str]
    """a message to show to students if this test cases fails"""


@dataclass
class TestCaseResult:
    """
    A dataclass representing the results of running a test case.
    """

    test_case: TestCase
    """the test case"""

    message: Optional[str]
    """an error message if the test case failed"""

    passed: bool
    """whether the test case was passed"""


class TestFile(ABC):
    """
    An (abstract) single test file for Otter. This ABC defines how test results are represented and
    sets up the instance variables tracked by tests. Subclasses should implement the abstract class
    method ``from_file`` and the abstract instance method ``run``.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list[TestCase]``): a list of parsed tests to be run
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across cases
    """

    name: str
    """the name of test file"""

    path: str
    """the path to the test file"""

    test_cases: list[TestCase]
    """a list of parsed tests to be run"""

    all_or_nothing: bool
    """whether the test should be graded all-or-nothing across cases"""

    test_case_results: list[TestCaseResult]
    """a list of results for the test cases in ``test_cases``"""

    _score: Optional[Union[int, float]]
    """an override for the overall score for this test file"""

    def _repr_html_(self):
        if self.passed_all:
            all_passed_emoji = random.choice(["🍀", "🎉", "🌈", "🙌", "🚀", "🌟", "✨", "💯"])
            if any(tcr.test_case.success_message is not None for tcr in self.test_case_results):
                ret = f"<p><strong><pre style='display: inline;'>{self.name}</pre></strong> passed! {all_passed_emoji}</p>"
                for tcr in self.test_case_results:
                    if tcr.test_case.success_message is not None:
                        ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.success_message}</p>"
                return ret
            return f"<p><strong><pre style='display: inline;'>{self.name}</pre></strong> passed! {all_passed_emoji}</p>"
        else:
            ret = f"<p><strong style='color: red;'><pre style='display: inline;'>{self.name}</pre> results:</strong></p>"
            for tcr in self.test_case_results:
                if tcr.passed and tcr.test_case.success_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.success_message}</p>"
                if not tcr.passed and tcr.test_case.failure_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.failure_message}</p>"
                ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> result:</strong></p>"
                ret += f"<pre>{indent(tcr.message or '', '    ')}</pre>"

            return ret

    def __repr__(self):
        return self.summary()

    def __init__(
        self, name: str, path: str, test_cases: list[TestCase], all_or_nothing: bool = True
    ):
        self.name = name
        self.path = path
        self.test_cases = test_cases
        self.all_or_nothing = all_or_nothing
        self.test_case_results = []
        self._score = None

    @staticmethod
    def resolve_test_file_points(
        total_points: Optional[Union[int, float, list[Union[int, float]]]],
        test_cases: list[TestCase],
    ) -> list[TestCase]:
        """
        Determine the point value for each test case in a question based on the specified question
        point total, if any, and a list of test cases.

        See :ref:`_test_files_python_resolve_point_values` for details on how point values are
        determined.

        Args:
            total_points (``int | float | list[int | float] | None``): the total point value for the
                question or a list of test case point values
            test_cases (``list[TestCase]``): the test cases in the question

        Returns:
            ``list[TestCase]``: a deep copy of ``test_cases`` with each test case's ``points`` field
                set to its resolved point value
        """
        if isinstance(total_points, list):
            if len(total_points) != len(test_cases):
                raise ValueError(
                    "Points specified in test has different length than number of test cases"
                )
            test_cases = [replace(tc, points=pt) for tc, pt in zip(test_cases, total_points)]
            total_points = None

        elif total_points is not None and not isinstance(total_points, (int, float)):
            raise TypeError(f"Test spec points has invalid type: {type(total_points)}")

        point_values = []
        for test_case in test_cases:
            if test_case.points is not None:
                assert type(test_case.points) in (
                    int,
                    float,
                ), f"Invalid point type: {type(test_case.points)}"
                point_values.append(test_case.points)

            else:
                point_values.append(None)

        pre_specified = sum(p for p in point_values if p is not None)
        if total_points is not None:
            if pre_specified > total_points:
                raise ValueError("More points specified in test cases than allowed for test")

            else:
                try:
                    per_remaining = (total_points - pre_specified) / sum(
                        1 for p in point_values if p is None
                    )
                except ZeroDivisionError:
                    per_remaining = 0.0

        else:
            if pre_specified == 0 and all(p in (0, None) for p in point_values):
                # if only zeros specified, assume test worth 1 pt and divide amongst nonzero cases
                try:
                    per_remaining = 1 / sum(p is None for p in point_values)
                except ZeroDivisionError:
                    per_remaining = 0.0

            else:
                # assume all other tests are worth 0 points
                per_remaining = 0.0

        point_values = [p if p is not None else per_remaining for p in point_values]
        return [replace(tc, points=p) for tc, p in zip(test_cases, point_values)]

    @property
    def passed_all(self) -> bool:
        """whether all test cases in this test file were passed"""
        return all(tcr.passed for tcr in self.test_case_results)

    @property
    def passed_all_public(self) -> bool:
        """whether all public test cases in this test file were passed"""
        return all(tcr.passed for tcr in self.test_case_results if not tcr.test_case.hidden)

    @property
    def all_public(self) -> bool:
        """whether all test cases in this test file are public"""
        return all(not tc.hidden for tc in self.test_cases)

    @property
    def grade(self) -> float:
        """the percentage of points earned in this test file"""
        if self.all_or_nothing and not self.passed_all:
            return 0
        elif self.all_or_nothing and self.passed_all:
            return 1
        else:
            return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed) / sum(
                tc.points for tc in self.test_cases
            )

    @property
    def score(self) -> Union[int, float]:
        """the total points earned in this test file"""
        if self._score is not None:
            return self._score
        if self.all_or_nothing:
            return self.grade * self.possible
        return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed)

    @property
    def possible(self) -> Union[int, float]:
        """the total points possible in this test file"""
        return sum(tc.points for tc in self.test_cases)

    def update_score(self, new_score: Union[int, float]):
        """
        Override the score for this test file to a specified value.

        Args:
            new_score (``int | float``): the new score
        """
        self._score = new_score

    def to_dict(self):
        """
        Convert this ``TestFile`` to a JSON-serializable ``dict``.

        Returns:
            ``dict[str, Any]``: the dictionary representation of this ``TestFile``
        """
        return {
            "score": self.score,
            "possible": self.possible,
            "name": self.name,
            "path": self.path,
            "test_cases": [asdict(tc) for tc in self.test_cases],
            "all_or_nothing": self.all_or_nothing,
            "test_case_results": [asdict(tcr) for tcr in self.test_case_results],
        }

    def summary(self, public_only: bool = False) -> str:
        """
        Generate a plaintext string summarizing the results of each test case in this test file.

        Args:
            public_only (``bool``): whether only public test cases should be included in the summary

        Returns:
            ``str``: the summary
        """
        if (not public_only and self.passed_all) or (public_only and self.passed_all_public):
            ret = f"{self.name} results: All test cases passed!"
            all_tcrs = self.test_case_results
            if public_only:
                all_tcrs = [tcr for tcr in self.test_case_results if not tcr.test_case.hidden]
            for tcr in all_tcrs:
                if tcr.test_case.success_message is not None:
                    ret += f"\n{tcr.test_case.name} message: {tcr.test_case.success_message}"
            return ret

        tcrs = self.test_case_results
        if public_only:
            tcrs = [tcr for tcr in tcrs if not tcr.test_case.hidden]

        tcr_summaries = []
        for tcr in tcrs:
            smry = ""
            if tcr.passed and tcr.test_case.success_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.success_message}\n\n"
            if not tcr.passed and tcr.test_case.failure_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.failure_message}\n\n"
            smry += f"{tcr.test_case.name} result:\n"
            smry += f"{indent((tcr.message or '').strip(), '    ')}\n\n"

            tcr_summaries.append(smry.strip())

        return f"{self.name} results:\n" + indent("\n\n".join(tcr_summaries), "    ")

    @classmethod
    @abstractmethod
    def from_file(cls, path: str) -> "TestFile":
        """
        Create a ``TestFile`` from the contents of the provided file path.

        Args:
            path (``str``): the test file path

        Returns:
            ``TestFile``: the initialized test file
        """
        ...

    @classmethod
    @abstractmethod
    def from_metadata(cls, s: Any, path: str) -> "TestFile":
        """
        Parse a test file from data stored in a notebook's metadata.

        Args:
            s (``Any``): the test file contents from the notebook metadata
            path (``str``): the path to the notebook

        Returns:
            ``TestFile``: the parsed test file.
        """
        ...

    @abstractmethod
    def run(self, global_environment: dict[str, Any]):
        """
        Run the test cases in this test file.

        Args:
            global_environment (``dict[str, Any]``): the environment to run the test cases against
        """
        ...
